import logging
import os
from uuid import NAMESPACE_OID, uuid5

from eventsourcing.application.process import ProcessApplication, WrappedRepository
from eventsourcing.domain.model.aggregate import BaseAggregateRoot
from eventsourcing.domain.model.decorators import retry
from eventsourcing.exceptions import OperationalError, RecordConflictError
from eventsourcing.tests.test_process import ExampleAggregate


def set_db_uri():
    host = os.getenv("MYSQL_HOST", "127.0.0.1")
    user = os.getenv("MYSQL_USER", "eventsourcing")
    password = os.getenv("MYSQL_PASSWORD", "eventsourcing")
    db_uri = (
        "mysql+pymysql://{}:{}@{}/eventsourcing?charset=utf8mb4&binary_prefix=true"
    ).format(user, password, host)
    os.environ["DB_URI"] = db_uri


class Order(BaseAggregateRoot):
    def __init__(self, **kwargs):
        super(Order, self).__init__(**kwargs)
        self.is_reserved = False
        self.is_paid = False

    class Event(BaseAggregateRoot.Event):
        pass

    class Created(Event, BaseAggregateRoot.Created):
        pass

    def set_is_reserved(self, reservation_id):
        self.__trigger_event__(Order.Reserved, reservation_id=reservation_id)

    class Reserved(Event):
        def mutate(self, order):
            assert not order.is_reserved, "Order {} already reserved.".format(order.id)
            order.is_reserved = True
            order.reservation_id = self.reservation_id

        @property
        def reservation_id(self):
            return self.__dict__["reservation_id"]

    def set_is_paid(self, payment_id):
        self.__trigger_event__(self.Paid, payment_id=payment_id)

    class Paid(Event):
        def mutate(self, order):
            assert not order.is_paid, "Order {} already paid.".format(order.id)
            order.is_paid = True
            order.payment_id = self.payment_id

        @property
        def payment_id(self):
            return self.__dict__["payment_id"]


class Reservation(BaseAggregateRoot):
    def __init__(self, order_id, **kwargs):
        super(Reservation, self).__init__(**kwargs)
        self.order_id = order_id

    class Event(BaseAggregateRoot.Event):
        pass

    class Created(Event, BaseAggregateRoot.Created):
        pass

    @classmethod
    def create(cls, order_id):
        return cls.__create__(
            originator_id=Reservation.create_reservation_id(order_id), order_id=order_id
        )

    @classmethod
    def create_reservation_id(cls, order_id):
        return uuid5(NAMESPACE_OID, str(order_id))


class Payment(BaseAggregateRoot):
    def __init__(self, order_id, **kwargs):
        super(Payment, self).__init__(**kwargs)
        self.order_id = order_id

    class Event(BaseAggregateRoot.Event):
        pass

    class Created(Event, BaseAggregateRoot.Created):
        pass

    @classmethod
    def make(self, order_id):
        return self.__create__(order_id=order_id)


@retry((OperationalError, RecordConflictError), max_attempts=10, wait=0.01)
def create_new_order():
    order = Order.__create__()
    order.__save__()
    return order.id


logger = logging.getLogger()


class Orders(ProcessApplication[Order, Order.Event]):
    persist_event_type = Order.Created
    # set_notification_ids = True
    # use_cache = True

    def policy(self, repository, event):
        if isinstance(event, Reservation.Created):
            # Set the order as reserved.
            order = self.get_order(order_id=event.order_id, repository=repository)
            assert not order.is_reserved
            order.set_is_reserved(event.originator_id)

        elif isinstance(event, Payment.Created):
            # Set the order as paid.
            order = repository[event.order_id]
            assert isinstance(order, Order)
            assert not order.is_paid
            order.set_is_paid(event.originator_id)

    @retry((OperationalError, RecordConflictError), max_attempts=10, wait=0.01)
    def create_new_order(self):
        order = Order.__create__()
        order.__save__()
        return order.id

    def is_order_reserved(self, order_id):
        order = self.get_order(order_id)
        return order is not None

    def is_order_paid(self, order_id):
        order = self.get_order(order_id)
        return order is not None and order.is_paid

    def get_order(self, order_id, repository: WrappedRepository = None):
        try:
            return (repository or self.repository)[order_id]
        except KeyError:
            return None


class Reservations(ProcessApplication[Reservation, Reservation.Event]):
    def policy(self, repository, event):
        if isinstance(event, Order.Created):
            # Create a reservation.
            return Reservation.create(order_id=event.originator_id)


class Payments(ProcessApplication):
    def policy(self, repository, event):
        if isinstance(event, Order.Reserved):
            # Make a payment.
            return Payment.make(order_id=event.originator_id)


class Examples(ProcessApplication[ExampleAggregate, ExampleAggregate.Event]):
    persist_event_type = ExampleAggregate.Created

    def policy(self, repository, event):
        if isinstance(event, ExampleAggregate.Created):
            example_aggregate = repository[event.originator_id]
            assert isinstance(example_aggregate, ExampleAggregate)
            example_aggregate.move_on()



import datetime
from time import sleep
from uuid import uuid4

from eventsourcing.domain.model.timebucketedlog import (
    Timebucketedlog,
    bucket_duration,
    bucket_starts,
    make_timebucket_id,
    start_new_timebucketedlog,
)
from eventsourcing.infrastructure.repositories.timebucketedlog_repo import (
    TimebucketedlogRepo,
)
from eventsourcing.infrastructure.timebucketedlog_reader import TimebucketedlogReader
from eventsourcing.tests.base import notquick
from eventsourcing.tests.sequenced_item_tests.base import WithEventPersistence
from eventsourcing.tests.sequenced_item_tests.test_cassandra_record_manager import (
    WithCassandraRecordManagers,
)
from eventsourcing.tests.sequenced_item_tests.test_sqlalchemy_record_manager import (
    SQLAlchemyRecordManagerTestCase,
)
from eventsourcing.utils.times import decimaltimestamp


class TimebucketedlogTestCase(WithEventPersistence):
    def test_entity_lifecycle(self):
        self.log_repo = TimebucketedlogRepo(self.entity_event_store)
        log_name = uuid4()
        log = self.log_repo.get_or_create(log_name=log_name, bucket_size="year")
        log = self.log_repo[log_name]
        self.assertIsInstance(log, Timebucketedlog)
        self.assertEqual(log.name, log_name)
        self.assertEqual(log.bucket_size, "year")

        # Append some messages to the log.
        message1 = "This is message 1"
        message2 = "This is message 2"
        message3 = "This is message 3"
        message4 = "This is message 4"
        message5 = "This is message 5"
        message6 = "This is message 6"
        event1 = log.log_message(message1)
        event2 = log.log_message(message2)
        event3 = log.log_message(message3)
        halfway = decimaltimestamp()
        event4 = log.log_message(message4)
        event5 = log.log_message(message5)
        event6 = log.log_message(message6)

        # Read messages from the log.
        reader = TimebucketedlogReader(log, event_store=self.log_event_store)
        messages = list(reader.get_messages(is_ascending=False))
        self.assertEqual(len(messages), 6)
        self.assertEqual(messages[0], message6)
        self.assertEqual(messages[1], message5)
        self.assertEqual(messages[2], message4)
        self.assertEqual(messages[3], message3)
        self.assertEqual(messages[4], message2)
        self.assertEqual(messages[5], message1)

        # Check we can get all the messages (query running in ascending order).
        messages = list(reader.get_messages(is_ascending=True))
        self.assertEqual(len(messages), 6)
        self.assertEqual(messages[0], message1)
        self.assertEqual(messages[1], message2)
        self.assertEqual(messages[2], message3)
        self.assertEqual(messages[3], message4)
        self.assertEqual(messages[4], message5)
        self.assertEqual(messages[5], message6)

        # Check we can get messages after halfway (query running in descending order).
        messages = list(reader.get_messages(gt=halfway, is_ascending=False))
        self.assertEqual(len(messages), 3)
        self.assertEqual(messages[0], message6)
        self.assertEqual(messages[1], message5)
        self.assertEqual(messages[2], message4)

        # Check we can get messages until halfway (query running in descending order).
        messages = list(reader.get_messages(lte=halfway, is_ascending=False))
        self.assertEqual(len(messages), 3)
        self.assertEqual(messages[0], message3)
        self.assertEqual(messages[1], message2)
        self.assertEqual(messages[2], message1)

        # Check we can get messages until halfway (query running in ascending order).
        messages = list(reader.get_messages(lte=halfway, is_ascending=True))
        self.assertEqual(len(messages), 3)
        self.assertEqual(messages[0], message1)
        self.assertEqual(messages[1], message2)
        self.assertEqual(messages[2], message3)

        # Check we can get messages after halfway (query running in ascending order).
        messages = list(reader.get_messages(gt=halfway, is_ascending=True))
        self.assertEqual(len(messages), 3)
        self.assertEqual(messages[0], message4)
        self.assertEqual(messages[1], message5)
        self.assertEqual(messages[2], message6)

        # Check we can get last three messages (query running in descending order).
        messages = list(reader.get_messages(limit=3, is_ascending=False))
        self.assertEqual(len(messages), 3)
        self.assertEqual(messages[0], message6)
        self.assertEqual(messages[1], message5)
        self.assertEqual(messages[2], message4)

        # Check we can get first three messages (query running in ascending order).
        messages = list(reader.get_messages(limit=3, is_ascending=True))
        self.assertEqual(len(messages), 3)
        self.assertEqual(messages[0], message1)
        self.assertEqual(messages[1], message2)
        self.assertEqual(messages[2], message3)

        # Check we can get last line (query running in descending order).
        messages = list(reader.get_messages(limit=1, gt=halfway, is_ascending=False))
        self.assertEqual(len(messages), 1)
        self.assertEqual(messages[0], message6)

        # Check we can get the first line after halfway (query running in ascending order).
        messages = list(reader.get_messages(limit=1, gt=halfway, is_ascending=True))
        self.assertEqual(len(messages), 1)
        self.assertEqual(messages[0], message4)

        # Check we can get the first line before halfway (query running in descending order).
        messages = list(reader.get_messages(limit=1, lte=halfway, is_ascending=False))
        self.assertEqual(len(messages), 1)
        self.assertEqual(messages[0], message3)

        # Check we can get the first line (query running in ascending order).
        messages = list(reader.get_messages(limit=1, lte=halfway, is_ascending=True))
        self.assertEqual(len(messages), 1)
        self.assertEqual(messages[0], message1)

        # Check there isn't a line after the last line (query running in ascending order).
        messages = list(
            reader.get_messages(limit=1, gt=event6.timestamp, is_ascending=True)
        )
        self.assertEqual(len(messages), 0)

        # Check there is nothing somehow both after and until halfway.
        messages = list(reader.get_messages(gt=halfway, lte=halfway))
        self.assertEqual(len(messages), 0)

    @notquick
    def test_buckets_size_second(self):
        # Start new log.
        log = start_new_timebucketedlog(name=uuid4(), bucket_size="second")

        # Write messages across the time interval
        start = datetime.datetime.now()
        number_of_messages = 300
        events = []
        for i in range(number_of_messages):
            message_logged = log.log_message(str(i))
            events.append(message_logged)
            sleep(0.01)
        self.assertGreater(
            datetime.datetime.now() - start, datetime.timedelta(seconds=1)
        )

        # Get the messages in descending order.
        reader = TimebucketedlogReader(log, self.log_event_store)
        messages = list(reader.get_messages(is_ascending=False, page_size=10))
        self.assertEqual(len(messages), number_of_messages)

        # Expect the order of the messages is the reverse of the created order.
        self.assertEqual(
            messages, list(reversed([str(i) for i in range(number_of_messages)]))
        )

        # Get the messages in ascending order.
        messages = list(reader.get_messages(is_ascending=True, page_size=10))
        self.assertEqual(len(messages), number_of_messages)

        # Expect the order of the messages is the same as the created order.
        self.assertEqual(messages, [str(i) for i in range(number_of_messages)])

        # Get a limited number of messages in descending order.
        limit = 150
        messages = list(
            reader.get_messages(is_ascending=False, page_size=10, limit=limit)
        )
        self.assertLess(limit, number_of_messages)
        self.assertEqual(len(messages), limit)

        # Expect the order of the messages is the reverse of the created order.
        self.assertEqual(
            messages,
            list(reversed([str(i) for i in range(number_of_messages)]))[:limit],
        )

        # Get a limited number of messages in ascending order.
        limit = 150
        messages = list(
            reader.get_messages(is_ascending=True, page_size=10, limit=limit)
        )
        self.assertLess(limit, number_of_messages)
        self.assertEqual(len(messages), limit)

        # Expect the order of the messages is the same as the created order.
        self.assertEqual(messages, [str(i) for i in range(limit)])

        # Get a limited number of messages in descending order from the midpoint down.
        limit = 110
        midpoint = events[150].timestamp
        messages = list(
            reader.get_messages(
                is_ascending=False, page_size=10, limit=limit, lt=midpoint
            )
        )
        self.assertLess(limit, number_of_messages)
        self.assertEqual(len(messages), limit)

        # Expect the order of the messages is the reverse of the created order.
        self.assertEqual(
            messages, list(reversed([str(i) for i in range(150 - limit, 150)]))
        )

        # Get a limited number of messages in ascending order from the midpoint up.
        limit = 110
        midpoint = events[149].timestamp
        messages = list(
            reader.get_messages(
                is_ascending=True, page_size=10, limit=limit, gt=midpoint
            )
        )
        self.assertLess(limit, number_of_messages)
        self.assertEqual(len(messages), limit)

        # Expect the order of the messages is the same as the created order.
        self.assertEqual(messages, [str(i) for i in range(150, 150 + limit)])

        # Get a limited number of messages in descending order above the midpoint down.
        limit = 200
        midpoint = events[150].timestamp
        messages = list(
            reader.get_messages(
                is_ascending=False, page_size=10, limit=limit, gte=midpoint
            )
        )
        self.assertLess(limit, number_of_messages)
        self.assertEqual(len(messages), 150)

        # Expect the order of the messages is the reverse of the created order.
        self.assertEqual(messages, list(reversed([str(i) for i in range(150, 300)])))

        # Get a limited number of messages in ascending order below the midpoint up.
        limit = 200
        midpoint = events[149].timestamp
        messages = list(
            reader.get_messages(
                is_ascending=True, page_size=10, limit=limit, lte=midpoint
            )
        )
        self.assertLess(limit, number_of_messages)
        self.assertEqual(len(messages), 150)

        # Expect the order of the messages is the same as the created order.
        self.assertEqual(messages, [str(i) for i in range(150)])

        #
        # Use the last position to start part way through.
        limit = 20
        last_position = reader.position
        messages = reader.get_messages(
            is_ascending=True, page_size=10, limit=limit, gt=last_position
        )
        messages = list(messages)
        self.assertEqual(len(messages), limit)

        # Expect the order of the messages is the same as the created order.
        self.assertEqual(messages, [str(i) for i in range(150, 150 + limit)])

        # Do it again.
        last_position = reader.position
        messages = reader.get_messages(
            is_ascending=True, page_size=10, limit=limit, gt=last_position
        )
        messages = list(messages)
        self.assertEqual(len(messages), limit)

        # Expect the order of the messages is the same as the created order.
        self.assertEqual(
            messages, [str(i) for i in range(150 + limit, 150 + limit * 2)]
        )

        # Go back.
        last_position = reader.position
        messages = reader.get_messages(
            is_ascending=False, page_size=10, limit=limit, lt=last_position
        )
        messages = list(messages)
        self.assertEqual(len(messages), limit)

        # Expect the order of the messages is the reverse of the created order.
        self.assertEqual(
            messages, [str(i) for i in range(148 + limit * 2, 148 + limit, -1)]
        )

        # Go back.
        last_position = reader.position
        messages = reader.get_messages(
            is_ascending=False, page_size=10, limit=limit, lt=last_position
        )
        messages = list(messages)
        self.assertEqual(len(messages), limit)

        # Expect the order of the messages is the reverse of the created order.
        self.assertEqual(
            messages, [str(i) for i in range(128 + limit * 2, 128 + limit, -1)]
        )

        # Go back.
        last_position = reader.position
        messages = reader.get_messages(
            is_ascending=False, page_size=10, limit=limit, lt=last_position
        )
        messages = list(messages)
        self.assertEqual(len(messages), limit)

        # Expect the order of the messages is the reverse of the created order.
        self.assertEqual(
            messages, [str(i) for i in range(108 + limit * 2, 108 + limit, -1)]
        )

        # Repeat.
        messages = reader.get_messages(
            is_ascending=False, page_size=10, limit=limit, lt=last_position
        )
        messages = list(messages)
        self.assertEqual(len(messages), limit)

        # Expect the order of the messages is the reverse of the created order.
        self.assertEqual(
            messages, [str(i) for i in range(108 + limit * 2, 108 + limit, -1)]
        )

    def test_buckets_of_all_sizes(self):
        # Start new second sized log.
        log_id2 = uuid4()
        log = start_new_timebucketedlog(name=log_id2, bucket_size="second")
        log.log_message("message")

        # Get the messages.
        reader = TimebucketedlogReader(log, self.log_event_store)
        self.assertTrue(len(list(reader.get_messages())))

        # Start new minute sized log.
        log_id3 = uuid4()
        log = start_new_timebucketedlog(name=log_id3, bucket_size="minute")
        log.log_message("message")

        # Get the messages.
        reader = TimebucketedlogReader(log, self.log_event_store)
        self.assertTrue(len(list(reader.get_messages())))

        # Start new hour sized log.
        log_id4 = uuid4()
        log = start_new_timebucketedlog(name=log_id4, bucket_size="hour")
        log.log_message("message")

        # Get the messages.
        reader = TimebucketedlogReader(log, self.log_event_store)
        self.assertTrue(len(list(reader.get_messages())))

        # Start new day sized log.
        log_id5 = uuid4()
        log = start_new_timebucketedlog(name=log_id5, bucket_size="day")
        log.log_message("message")

        # Get the messages.
        reader = TimebucketedlogReader(log, self.log_event_store)
        self.assertTrue(len(list(reader.get_messages())))

        # Start new month sized log.
        log_id6 = uuid4()
        log = start_new_timebucketedlog(name=log_id6, bucket_size="month")
        log.log_message("message")

        # Get the messages.
        reader = TimebucketedlogReader(log, self.log_event_store)
        self.assertTrue(len(list(reader.get_messages())))

        # Start new year sized log.
        log_id7 = uuid4()
        log = start_new_timebucketedlog(name=log_id7, bucket_size="year")
        log.log_message("message")

        # Get the messages.
        reader = TimebucketedlogReader(log, self.log_event_store)
        self.assertTrue(len(list(reader.get_messages())))

        # Start new default sized log.
        log_id8 = uuid4()
        log = start_new_timebucketedlog(name=log_id8)
        log.log_message("message")

        # Get the messages.
        reader = TimebucketedlogReader(log, self.log_event_store)
        self.assertTrue(len(list(reader.get_messages())))

        # Start new invalid sized log.
        with self.assertRaises(ValueError):
            log_id9 = uuid4()
            log = start_new_timebucketedlog(name=log_id9, bucket_size="invalid")

        # Check the helper methods are protected against invalid bucket sizes.
        with self.assertRaises(ValueError):
            log_id10 = uuid4()
            make_timebucket_id(log_id10, decimaltimestamp(), bucket_size="invalid")

        with self.assertRaises(ValueError):
            bucket_starts(decimaltimestamp(), bucket_size="invalid")

        with self.assertRaises(ValueError):
            bucket_duration(bucket_size="invalid")


class TestLogWithCassandra(WithCassandraRecordManagers, TimebucketedlogTestCase):
    pass


class TestLogWithSQLAlchemy(SQLAlchemyRecordManagerTestCase, TimebucketedlogTestCase):
    pass


del TimebucketedlogTestCase
del WithEventPersistence



from functools import _find_impl, singledispatch, wraps
from inspect import isfunction
from typing import Callable, no_type_check


def applicationpolicy(arg: Callable) -> Callable:
    """
    Decorator for application policy method.

    Allows policy to be built up from methods
    registered for different event classes.
    """

    assert isfunction(arg), arg

    @no_type_check
    def _mutator(func):
        wrapped = singledispatch(func)

        @wraps(wrapped)
        def wrapper(*args, **kwargs):
            event = kwargs.get("event") or args[-1]
            return wrapped.dispatch(type(event))(*args, **kwargs)

        wrapper.register = wrapped.register

        return wrapper

    return _mutator(arg)


def applicationpolicy2(arg: Callable) -> Callable:
    """
    This one doesn't use weakrefs.
    """
    handlers = {}
    cache = {}

    def _mutator(func):
        def dispatch(event_type):
            try:
                return cache[event_type]
            except KeyError:
                handler = _find_impl(event_type, handlers) or func
                cache[event_type] = handler
                return handler

        @wraps(func)
        def policy_function_wrapper(*args, **kwargs):
            event = kwargs.get("event") or args[-1]
            return dispatch(type(event))(*args, **kwargs)

        def register(event_type):
            def registered_function_decorator(func):
                handlers[event_type] = func
                return func

            return registered_function_decorator

        policy_function_wrapper.register = register

        return policy_function_wrapper

    return _mutator(arg)



import os
import zlib
from json import JSONDecoder, JSONEncoder
from typing import (
    Any,
    Dict,
    Generic,
    Iterable,
    List,
    NamedTuple,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
)

from eventsourcing.application.notificationlog import (
    LocalNotificationLog,
    RecordManagerNotificationLog,
)
from eventsourcing.application.pipeline import Pipeable
from eventsourcing.application.policies import PersistencePolicy
from eventsourcing.domain.model.aggregate import BaseAggregateRoot, TAggregateEvent
from eventsourcing.domain.model.entity import (
    TDomainEvent,
    TVersionedEntity,
    TVersionedEvent,
)
from eventsourcing.domain.model.events import DomainEvent, publish
from eventsourcing.exceptions import ProgrammingError, PromptFailed
from eventsourcing.infrastructure.base import (
    DEFAULT_PIPELINE_ID,
    AbstractEventStore,
    AbstractRecordManager,
    BaseRecordManager,
    RecordManagerWithNotifications,
    RecordManagerWithTracking,
    TrackingKwargs,
)
from eventsourcing.infrastructure.datastore import AbstractDatastore
from eventsourcing.infrastructure.eventsourcedrepository import EventSourcedRepository
from eventsourcing.infrastructure.eventstore import EventStore
from eventsourcing.infrastructure.factory import InfrastructureFactory
from eventsourcing.infrastructure.sequenceditem import StoredEvent
from eventsourcing.infrastructure.sequenceditemmapper import SequencedItemMapper
from eventsourcing.utils.cipher.aes import AESCipher
from eventsourcing.utils.random import decode_bytes
from eventsourcing.whitehead import ActualOccasion, IterableOfEvents, T

PersistEventType = Optional[Union[Type[DomainEvent], Tuple[Type[DomainEvent]]]]

CausalDependencies = Dict[str, int]
ListOfCausalDependencies = List[CausalDependencies]


class ProcessEvent(ActualOccasion, Generic[TDomainEvent]):
    def __init__(
        self,
        domain_events: Iterable[TDomainEvent],
        tracking_kwargs: Optional[TrackingKwargs] = None,
        causal_dependencies: Optional[ListOfCausalDependencies] = None,
        orm_objs_pending_save: Sequence[Any] = (),
        orm_objs_pending_delete: Sequence[Any] = (),
    ):
        self.domain_events = domain_events
        self.tracking_kwargs = tracking_kwargs
        self.causal_dependencies = causal_dependencies
        self.orm_objs_pending_save = orm_objs_pending_save
        self.orm_objs_pending_delete = orm_objs_pending_delete


class SimpleApplication(Pipeable, Generic[TVersionedEntity, TVersionedEvent]):
    """
    Base class for event sourced applications.

    Constructs infrastructure objects such as the repository and
    event store, and also the notification log which presents the
    application state as a sequence of events.

    Needs actual infrastructure classes.
    """

    infrastructure_factory_class: Type[InfrastructureFactory] = InfrastructureFactory
    is_constructed_with_session: bool = False

    record_manager_class: Optional[Type[AbstractRecordManager]] = None
    stored_event_record_class: Optional[type] = None
    snapshot_record_class: Optional[type] = None

    sequenced_item_class: Optional[Type[NamedTuple]] = None
    sequenced_item_mapper_class: Optional[Type[SequencedItemMapper]] = None
    compressor: Any = None
    json_encoder_class: Optional[Type[JSONEncoder]] = None
    sort_keys: bool = False
    json_decoder_class: Optional[Type[JSONDecoder]] = None

    persist_event_type: Optional[PersistEventType] = None
    notification_log_section_size: Optional[int] = None
    use_cache: bool = False

    event_store_class: Type[EventStore] = EventStore
    repository_class: Type[EventSourcedRepository] = EventSourcedRepository

    use_causal_dependencies = False
    set_notification_ids = False

    def __init__(
        self,
        name: str = "",
        persistence_policy: Optional[PersistencePolicy] = None,
        persist_event_type: PersistEventType = None,
        cipher_key: Optional[str] = None,
        compressor: Any = None,
        sequenced_item_class: Optional[Type[NamedTuple]] = None,
        sequenced_item_mapper_class: Optional[Type[SequencedItemMapper]] = None,
        record_manager_class: Optional[Type[AbstractRecordManager]] = None,
        stored_event_record_class: Optional[type] = None,
        event_store_class: Optional[Type[EventStore]] = None,
        snapshot_record_class: Optional[type] = None,
        setup_table: bool = True,
        contiguous_record_ids: bool = True,
        pipeline_id: int = DEFAULT_PIPELINE_ID,
        json_encoder_class: Optional[Type[JSONEncoder]] = None,
        sort_keys: bool = False,
        json_decoder_class: Optional[Type[JSONDecoder]] = None,
        notification_log_section_size: Optional[int] = None,
        use_cache: bool = False,
    ):
        """
        Initialises application object.

        :param name: Name of application.
        :param persistence_policy: Persistence policy object.
        :param persist_event_type: Tuple of domain event classes to be persisted.
        :param cipher_key: Base64 unicode string cipher key.
        :param compressor: Compressor used to compress serialized event state.
        :param sequenced_item_class: Named tuple for mapping and recording events.
        :param sequenced_item_mapper_class: Object class for mapping stored events.
        :param record_manager_class: Object class for recording stored events.
        :param stored_event_record_class: Object class for event records.
        :param event_store_class: Object class uses to store and retrieve domain events.
        :param snapshot_record_class: Object class used to represent snapshots.
        :param setup_table: Option to create database tables when application starts.
        :param contiguous_record_ids: Whether or not to delegate notification ID
            generation to the record manager (to guarantee there will be no gaps).
        :param pipeline_id: ID of instance of system pipeline expressions.
        :param json_encoder_class: Object class used to encode object as JSON strings.
        :param json_decoder_class: Object class used to decode JSON strings as objects.
        :param notification_log_section_size: Number of notification items in a section.
        :param use_cache: Whether or not to keep aggregates in memory (saves replaying
            when accessing again, but uses memory).
        """
        self.name = name or type(self).create_name()

        self.notification_log_section_size = (
            notification_log_section_size or type(self).notification_log_section_size
        )

        sequenced_item_class = sequenced_item_class or type(self).sequenced_item_class
        sequenced_item_class = sequenced_item_class or StoredEvent  # type: ignore
        self.sequenced_item_class = sequenced_item_class
        assert self.sequenced_item_class is not None
        self.sequenced_item_mapper_class: Type[SequencedItemMapper] = (
            sequenced_item_mapper_class
            or type(self).sequenced_item_mapper_class
            or SequencedItemMapper
        )

        self.record_manager_class = (
            record_manager_class or type(self).record_manager_class
        )
        self._stored_event_record_class = stored_event_record_class
        self._snapshot_record_class = snapshot_record_class

        self.event_store_class = event_store_class or type(self).event_store_class

        self.json_encoder_class = json_encoder_class or type(self).json_encoder_class
        self.sort_keys = sort_keys or type(self).sort_keys
        self.json_decoder_class = json_decoder_class or type(self).json_decoder_class
        self.persist_event_type = persist_event_type or type(self).persist_event_type

        self.contiguous_record_ids = contiguous_record_ids
        self.pipeline_id = pipeline_id
        self._persistence_policy = persistence_policy

        self.cipher = self.construct_cipher(cipher_key)
        self.compressor = compressor or type(self).compressor

        # Default to using zlib compression when encrypting.
        if self.cipher and self.compressor is None:
            self.compressor = zlib

        self.infrastructure_factory: Optional[
            InfrastructureFactory[TVersionedEvent]
        ] = None
        self._datastore: Optional[AbstractDatastore] = None
        self._event_store: Optional[
            AbstractEventStore[TVersionedEvent, BaseRecordManager]
        ] = None
        self._repository: Optional[
            EventSourcedRepository[TVersionedEntity, TVersionedEvent]
        ] = None
        self._notification_log: Optional[LocalNotificationLog] = None

        self.use_cache = use_cache or type(self).use_cache

        if (
            self.record_manager_class
            or self.infrastructure_factory_class.record_manager_class
        ):

            self.construct_infrastructure()

            if setup_table:
                self.setup_table()
            self.construct_notification_log()

            if self._persistence_policy is None:
                self.construct_persistence_policy()

    @classmethod
    def create_name(cls):
        return cls.__name__.lower()

    @property
    def datastore(self) -> AbstractDatastore:
        if self._datastore is None:
            self._raise_on_missing_infrastructure("datastore")
        return self._datastore

    @property
    def event_store(self) -> AbstractEventStore[TVersionedEvent, BaseRecordManager]:
        if self._event_store is None:
            self._raise_on_missing_infrastructure("event_store")
        return self._event_store

    @property
    def repository(self) -> EventSourcedRepository[TVersionedEntity, TVersionedEvent]:
        if self._repository is None:
            self._raise_on_missing_infrastructure("repository")
        return self._repository

    @property
    def notification_log(self) -> LocalNotificationLog:
        if self._notification_log is None:
            self._raise_on_missing_infrastructure("notification_log")
        return self._notification_log

    @property
    def persistence_policy(self) -> PersistencePolicy:
        if self._persistence_policy is None:
            self._raise_on_missing_infrastructure("persistence_policy")
        return self._persistence_policy

    def _raise_on_missing_infrastructure(self, what_is_missing):
        if not isinstance(self, ApplicationWithConcreteInfrastructure):
            msg = (
                "Application class %s is not an subclass of %s."
                " Try using or inheriting from or mixin() an application"
                " class with concrete infrastructure such as SQLAlchemyApplication"
                " or DjangoApplication or AxonApplication."
            ) % (type(self), ApplicationWithConcreteInfrastructure)
        else:
            msg = "Application class %s does not have a %s" % (
                type(self).__name__,
                what_is_missing,
            )
        raise ProgrammingError(msg)

    def construct_cipher(self, cipher_key_str: Optional[str]) -> Optional[AESCipher]:
        cipher_key_bytes = decode_bytes(
            cipher_key_str or os.getenv("CIPHER_KEY", "") or ""
        )
        return AESCipher(cipher_key_bytes) if cipher_key_bytes else None

    def construct_infrastructure(self, *args: Any, **kwargs: Any) -> None:
        """
        Constructs infrastructure for application.
        """
        self.infrastructure_factory = self.construct_infrastructure_factory(
            *args, **kwargs
        )
        self.construct_datastore()
        self.construct_event_store()
        self.construct_repository()

    def construct_infrastructure_factory(
        self, *args: Any, **kwargs: Any
    ) -> InfrastructureFactory:
        """
        Constructs infrastructure factory object.
        """
        factory_class = self.infrastructure_factory_class
        assert issubclass(factory_class, InfrastructureFactory)

        integer_sequenced_record_class = (
            self._stored_event_record_class or self.stored_event_record_class
        )
        snapshot_record_class = (
            self._snapshot_record_class or self.snapshot_record_class
        )

        return factory_class(  # type:ignore  # multiple values for keyword argument
            record_manager_class=self.record_manager_class,
            integer_sequenced_record_class=integer_sequenced_record_class,
            snapshot_record_class=snapshot_record_class,
            sequenced_item_class=self.sequenced_item_class,
            sequenced_item_mapper_class=self.sequenced_item_mapper_class,
            json_encoder_class=self.json_encoder_class,
            sort_keys=self.sort_keys,
            json_decoder_class=self.json_decoder_class,
            contiguous_record_ids=self.contiguous_record_ids,
            application_name=self.name,
            pipeline_id=self.pipeline_id,
            event_store_class=self.event_store_class,
            *args,
            **kwargs
        )

    def construct_datastore(self) -> None:
        """
        Constructs datastore object (used to create and drop database tables).
        """
        assert self.infrastructure_factory
        self._datastore = self.infrastructure_factory.construct_datastore()

    def construct_event_store(self) -> None:
        """
        Constructs event store object.
        """
        assert self.infrastructure_factory
        factory = self.infrastructure_factory
        self._event_store = factory.construct_integer_sequenced_event_store(
            self.cipher, self.compressor
        )

    def construct_repository(self, **kwargs: Any) -> None:
        """
        Constructs repository object.
        """
        assert self.repository_class
        self._repository = self.repository_class(
            event_store=self.event_store, use_cache=self.use_cache, **kwargs
        )

    def setup_table(self) -> None:
        """
        Sets up the database table using event store's record class.
        """
        if self._datastore is not None:
            record_class = self.event_store.record_manager.record_class
            self.datastore.setup_table(record_class)

    def drop_table(self) -> None:
        """
        Drops the database table using event store's record class.
        """
        if self._datastore is not None:
            record_class = self.event_store.record_manager.record_class
            self.datastore.drop_table(record_class)

    def construct_notification_log(self) -> None:
        """
        Constructs notification log object.
        """
        self._notification_log = RecordManagerNotificationLog(
            self.event_store.record_manager,
            section_size=self.notification_log_section_size,
        )

    def construct_persistence_policy(self) -> None:
        """
        Constructs persistence policy object.
        """
        self._persistence_policy = PersistencePolicy(
            event_store=self.event_store, persist_event_type=self.persist_event_type
        )

    def change_pipeline(self, pipeline_id: int) -> None:
        """
        Switches pipeline being used by this application object.
        """
        self.pipeline_id = pipeline_id
        self.event_store.record_manager.pipeline_id = pipeline_id

    def close(self) -> None:
        """
        Closes the application for further use.

        The persistence policy is closed, and the application's
        connection to the database is closed.
        """
        # Close the persistence policy.
        if self._persistence_policy is not None:
            self._persistence_policy.close()

        # Close database connection.
        if self._datastore is not None:
            self._datastore.close_connection()

    def __enter__(self: T) -> T:
        """
        Supports use of application as context manager.
        """
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        """
        Closes application when exiting context manager.
        """
        self.close()

    @classmethod
    def reset_connection_after_forking(cls) -> None:
        pass

    @classmethod
    def mixin(cls: T, infrastructure_class: type) -> T:
        """
        Returns subclass that inherits also from given infrastructure class.
        """
        return type(cls.__name__, (infrastructure_class, cls), {})

    def save(
        self, aggregates=(), orm_objects_pending_save=(), orm_objects_pending_delete=(),
    ) -> None:
        """
        Saves state of aggregates, and ORM objects.

        All of the pending events of the aggregates, along with the
        ORM objects, are recorded atomically as a process event.

        Then a "prompt to pull" is published, and, if the repository cache
        is in use, then puts the aggregates in the cache.

        :param aggregates: One or many aggregates.
        :param orm_objects_pending_save: Sequence of ORM objects to be saved.
        :param orm_objects_pending_delete: Sequance of ORM objects to be deleted.
        """
        # Collect pending events from the aggregates.
        new_events = []
        if isinstance(aggregates, BaseAggregateRoot):
            aggregates = [aggregates]
        else:
            pass
            # Todo: Make sure record manager supports writing events from more than
            #  one aggregate atomically (e.g. Cassandra and EventStore don't).

        for aggregate in aggregates:
            new_events += aggregate.__batch_pending_events__()
        process_event = ProcessEvent(
            domain_events=new_events,
            orm_objs_pending_save=orm_objects_pending_save,
            orm_objs_pending_delete=orm_objects_pending_delete,
        )
        new_records = self.record_process_event(process_event)
        record_manager = self.event_store.record_manager
        if isinstance(record_manager, RecordManagerWithNotifications):
            # Find the head notification ID.
            notifiable_events = [e for e in new_events if e.__notifiable__]
            head_notification_id = None
            if len(notifiable_events):
                notification_id_name = record_manager.notification_id_name
                notifications = []
                for record in new_records:
                    if not hasattr(record, notification_id_name):
                        continue
                    if not isinstance(getattr(record, notification_id_name), int):
                        continue
                    notifications.append(
                        record_manager.create_notification_from_record(record)
                    )

                if len(notifications):
                    head_notification_id = notifications[-1]["id"]
            self.publish_prompt(head_notification_id)
        if self.repository.use_cache:
            for aggregate in aggregates:
                self.repository.put_entity_in_cache(aggregate.id, aggregate)

    def record_process_event(self, process_event: ProcessEvent) -> List:
        """
        Records a process event.

        Converts the domain events of the process event to event record
        objects, and writes the event records and the ORM objects to
        the database using the application's event store's record manager.

        :param process_event: An instance of
            :class:`~eventsourcing.application.simple.ProcessEvent`
        :return: A list of event records.
        """
        # Construct event records.
        event_records = self.construct_event_records(
            process_event.domain_events, process_event.causal_dependencies
        )

        # Write event records with tracking record.
        record_manager = self.event_store.record_manager
        assert isinstance(record_manager, RecordManagerWithTracking)
        record_manager.write_records(
            records=event_records,
            tracking_kwargs=process_event.tracking_kwargs,
            orm_objs_pending_save=process_event.orm_objs_pending_save,
            orm_objs_pending_delete=process_event.orm_objs_pending_delete,
        )
        return event_records

    def construct_event_records(
        self,
        pending_events: Iterable[TAggregateEvent],
        causal_dependencies: Optional[ListOfCausalDependencies],
    ) -> List:
        """
        Constructs event records from domain events.

        :param pending_events: An iterable of domain events.
        :param causal_dependencies: A list of causal dependencies.
        :return: A list of event records.
        """
        # Convert to event records.
        sequenced_items = self.event_store.items_from_events(pending_events)
        record_manager = self.event_store.record_manager
        assert record_manager
        assert isinstance(record_manager, RecordManagerWithTracking)
        event_records = list(record_manager.to_records(sequenced_items))

        # Set notification log IDs, and causal dependencies.
        if len(event_records):
            # Todo: Maybe keep track of what this probably is, to
            #  avoid query. Like log reader, invalidate on error.
            if self.set_notification_ids:
                notification_id_name = record_manager.notification_id_name
                current_max = record_manager.get_max_notification_id()
                for domain_event, event_record in zip(pending_events, event_records):
                    if type(domain_event).__notifiable__:
                        current_max += 1
                        setattr(event_record, notification_id_name, current_max)
                    else:
                        setattr(
                            event_record, notification_id_name, "event-not-notifiable"
                        )
            else:
                if any((not e.__notifiable__ for e in pending_events)):
                    raise Exception(
                        "Can't set __notifiable__=False without "
                        "set_notification_ids=True "
                    )

            if self.use_causal_dependencies:
                assert hasattr(record_manager.record_class, "causal_dependencies")
                causal_dependencies_json = self.event_store.event_mapper.json_dumps(
                    causal_dependencies
                ).decode("utf8")
                # Only need first event to carry the dependencies.
                event_records[0].causal_dependencies = causal_dependencies_json

        return event_records

    def publish_prompt(self, head_notification_id=None):
        """
        Publishes a "prompt to pull" (instance of
        :class:`~eventsourcing.application.simple.PromptToPull`).

        :param head_notification_id: Maximum notification ID of event records
            to be pulled.
        """
        prompt = PromptToPull(self.name, self.pipeline_id, head_notification_id)
        try:
            publish(prompt)
        except PromptFailed:
            raise
        except Exception as e:
            raise PromptFailed("{}: {}".format(type(e), str(e)))


class ApplicationWithConcreteInfrastructure(SimpleApplication):
    """
    Base class for application classes that have actual infrastructure.
    """


class Prompt(ActualOccasion):
    pass


def is_prompt_to_pull(events: IterableOfEvents) -> bool:
    return isinstance(events, PromptToPull)


class PromptToPull(Prompt):
    def __init__(self, process_name: str, pipeline_id: int, head_notification_id=None):
        self.process_name: str = process_name
        self.pipeline_id: int = pipeline_id
        self.head_notification_id = head_notification_id

    def __eq__(self, other: object) -> bool:
        return bool(
            other
            and isinstance(other, type(self))
            and self.process_name == other.process_name
            and self.pipeline_id == other.pipeline_id
        )

    def __repr__(self) -> str:
        return "{}({}={}, {}={})".format(
            type(self).__name__,
            "process_name",
            self.process_name,
            "pipeline_id",
            self.pipeline_id,
        )



import os
from base64 import b64decode, b64encode


def encoded_random_bytes(num_bytes: int) -> str:
    """
    Generates random bytes, encoded as Base64 unicode string.

    :param num_bytes: Number of random bytes to generate.
    :return: Random bytes of specified length, encoded as Base64 unicode string.
    """
    return encode_bytes(random_bytes(num_bytes))


encode_random_bytes = encoded_random_bytes  # Backward compatibility with <= 7.2.9999.


def random_bytes(num_bytes: int) -> bytes:
    """
    Generates random bytes.

    :param num_bytes: Number of random bytes to generate.
    :return: Random bytes of specified length.
    :type: bytes
    """
    return os.urandom(num_bytes)


def encode_bytes(value: bytes) -> str:
    """
    Encodes value as Base64 unicode string.

    :param value: Bytes to be encoded.
    """
    return b64encode(value).decode("utf-8")


def decode_bytes(value: str) -> bytes:
    """
    Decodes bytes from Base64 encoded unicode string.

    :param str value: Base64 string.
    :returns: Bytes that were encoded with Base64.
    :rtype: bytes
    """
    return b64decode(value.encode("utf-8"))



import os
import unittest
from time import sleep

import django
from django.core.management import call_command
from django.test import TransactionTestCase

from eventsourcing.infrastructure.django.apps import DjangoConfig
from eventsourcing.infrastructure.django.factory import DjangoInfrastructureFactory
from eventsourcing.infrastructure.django.utils import close_django_connection
from eventsourcing.tests.sequenced_item_tests import base

os.environ[
    "DJANGO_SETTINGS_MODULE"
] = "eventsourcing.tests.djangoproject.djangoproject.settings"

django.setup()


class DjangoTestCase(TransactionTestCase):
    infrastructure_factory_class = DjangoInfrastructureFactory
    contiguous_record_ids = True

    def setUp(self):
        super(DjangoTestCase, self).setUp()
        # Setup tables (there isn't a Django datastore object, but we can do it like
        # this).
        call_command("migrate", verbosity=0, interactive=False)
        sleep(1)

    def tearDown(self):
        # Drop tables (there isn't a Django datastore object, but we can do it like
        # this).
        # call_command('migrate', 'django', 'zero', verbosity=0, interactive=False)
        super(DjangoTestCase, self).tearDown()

    def construct_datastore(self):
        pass

    def close_connections_before_forking(self):
        # If connection is already made close it.
        close_django_connection()


class TestDjangoRecordManagerWithIntegerSequences(
    DjangoTestCase, base.IntegerSequencedRecordTestCase
):
    def construct_record_manager(self):
        return self.construct_entity_record_manager()


class TestDjangoRecordManagerWithoutContiguousRecordIDs(
    DjangoTestCase, base.IntegerSequencedRecordTestCase
):
    contiguous_record_ids = False

    def construct_record_manager(self):
        return self.construct_entity_record_manager()


class TestDjangoRecordManagerWithTimestampSequences(
    DjangoTestCase, base.TimestampSequencedItemTestCase
):
    def construct_record_manager(self):
        return self.construct_timestamp_sequenced_record_manager()


# class WithDjangoRecordManagers(DjangoTestCase, WithActiveRecordManagers):
#     def construct_entity_record_manager(self):
#         return construct_integer_sequenced_record_manager()
#
#     def construct_log_record_manager(self):
#         return construct_timestamp_sequenced_record_manager()
#
#     def construct_snapshot_record_manager(self):
#         return construct_snapshot_record_manager()
#

# class TestSimpleSequencedItemIteratorWithDjango(WithDjangoRecordManagers,
#                                                 SimpleSequencedItemteratorTestCase):
#     pass
#
#
# class TestThreadedSequencedItemIteratorWithDjango(WithDjangoRecordManagers,
#                                                   ThreadedSequencedItemIteratorTestCase):
#     pass


class TestConfigClass(unittest.TestCase):
    def test(self):
        self.assertEqual("eventsourcing.infrastructure.django", DjangoConfig.name)



import os
from typing import Any, Optional

from axonclient.client import DEFAULT_LOCAL_AXONSERVER_URI, AxonClient

from eventsourcing.infrastructure.datastore import AbstractDatastore, DatastoreSettings


class AxonSettings(DatastoreSettings):
    def __init__(self, uri: Optional[str] = None):
        if uri is not None:
            self.uri = uri
        else:
            if "AXON_HOST" in os.environ:
                axon_db_uri = "{}:{}".format(
                    os.getenv("AXON_HOST"), os.getenv("AXON_PORT", "8124")
                )
            else:
                axon_db_uri = DEFAULT_LOCAL_AXONSERVER_URI
            self.uri = os.getenv("DB_URI", axon_db_uri)


class AxonDatastore(AbstractDatastore[AxonSettings]):
    can_drop_tables = False

    def __init__(self, settings: AxonSettings):
        super(AxonDatastore, self).__init__(settings=settings)
        self._axon_client: Optional[AxonClient] = None

    @property
    def axon_client(self):
        if self._axon_client is None:
            self.setup_connection()
        return self._axon_client

    def setup_connection(self) -> None:
        assert isinstance(self.settings, AxonSettings), self.settings
        if self._axon_client is None:
            self._axon_client = AxonClient(self.settings.uri)

    def close_connection(self):
        if self.axon_client is not None:
            self.axon_client.close_connection()

    def drop_table(self, table: Any) -> None:
        pass

    def drop_tables(self) -> None:
        pass

    def setup_table(self, table: Any) -> None:
        pass

    def setup_tables(self) -> None:
        pass

    def truncate_tables(self) -> None:
        pass



from unittest import skip

from eventsourcing.application.django import DjangoApplication
from eventsourcing.tests.sequenced_item_tests.test_django_record_manager import (
    DjangoTestCase,
)
from eventsourcing.tests.test_ray_runner import TestRayRunner


@skip("This doesn't work at the moment (same forking/setup problem?)")
class DontTestRayRunnerWithDjango(DjangoTestCase, TestRayRunner):
    infrastructure_class = DjangoApplication


# Don't let this be found here.
del TestRayRunner



from typing import Any, Dict, Tuple, Type

from eventsourcing.infrastructure.sequenceditemmapper import SequencedItemMapper
from eventsourcing.utils.topic import get_topic, resolve_topic
from eventsourcing.whitehead import TEvent


class SequencedItemMapperForPopo(SequencedItemMapper):
    def get_event_class_and_attrs(
        self, topic: str, state: bytes
    ) -> Tuple[Type[TEvent], Dict]:
        return resolve_topic(topic), state  # type: ignore

    def get_item_topic_and_state(
        self, domain_event_class: type, event_attrs: Dict[str, Any]
    ) -> Tuple[str, bytes]:
        return get_topic(domain_event_class), event_attrs  # type: ignore



import time
from decimal import Decimal
from typing import Iterable, Optional

from eventsourcing.domain.model.timebucketedlog import (
    MessageLogged,
    Timebucketedlog,
    make_timebucket_id,
    next_bucket_starts,
    previous_bucket_starts,
)
from eventsourcing.infrastructure.base import AbstractEventStore


def get_timebucketedlog_reader(
    log: Timebucketedlog, event_store: AbstractEventStore
) -> "TimebucketedlogReader":
    return TimebucketedlogReader(log=log, event_store=event_store)


class TimebucketedlogReader(object):
    def __init__(
        self, log: Timebucketedlog, event_store: AbstractEventStore, page_size: int = 50
    ):
        self.log = log
        self.event_store = event_store
        self.page_size = page_size
        self.position: Optional[Decimal] = None

    def get_messages(
        self,
        gt: Optional[int] = None,
        gte: Optional[int] = None,
        lt: Optional[int] = None,
        lte: Optional[int] = None,
        limit: Optional[int] = None,
        is_ascending: bool = False,
        page_size: Optional[int] = None,
    ) -> Iterable[str]:
        events = self.get_events(
            gt=gt,
            gte=gte,
            lt=lt,
            lte=lte,
            limit=limit,
            is_ascending=is_ascending,
            page_size=page_size,
        )
        for event in events:
            if isinstance(event, MessageLogged):
                self.position = event.timestamp
                yield event.message

    def get_events(
        self,
        gt: Optional[int] = None,
        gte: Optional[int] = None,
        lt: Optional[int] = None,
        lte: Optional[int] = None,
        limit: Optional[int] = None,
        is_ascending: bool = False,
        page_size: Optional[int] = None,
    ) -> Iterable[MessageLogged]:
        assert limit is None or limit > 0

        # Identify the first time bucket.
        now = time.time()
        started_on = self.log.started_on
        absolute_latest = min(float(now), lt or now, lte or now)
        absolute_earlyist = max(float(started_on), gt or 0, gte or 0)
        if is_ascending:
            position = absolute_earlyist
        else:
            position = absolute_latest

        # Start counting events.
        count_events = 0

        while True:
            bucket_id = make_timebucket_id(
                self.log.name, position, self.log.bucket_size
            )
            for message_logged_event in self.event_store.iter_events(
                originator_id=bucket_id,
                gt=gt,
                gte=gte,
                lt=lt,
                lte=lte,
                limit=limit,
                is_ascending=is_ascending,
                page_size=page_size,
            ):
                yield message_logged_event

                if limit is not None:
                    count_events += 1
                    if count_events >= limit:
                        return

            # See if there's another bucket.
            if is_ascending:
                next_timestamp = next_bucket_starts(position, self.log.bucket_size)
                if next_timestamp > absolute_latest:
                    return
                else:
                    position = next_timestamp
            else:
                if position < absolute_earlyist:
                    return
                else:
                    position = previous_bucket_starts(position, self.log.bucket_size)



from eventsourcing.application.simple import ApplicationWithConcreteInfrastructure
from eventsourcing.infrastructure.axonserver.factory import AxonInfrastructureFactory


class AxonApplication(ApplicationWithConcreteInfrastructure):
    infrastructure_factory_class = AxonInfrastructureFactory



from unittest import TestCase

from eventsourcing.application.axon import AxonApplication
from eventsourcing.application.django import DjangoApplication
from eventsourcing.application.notificationlog import NotificationLogReader
from eventsourcing.application.popo import PopoApplication
from eventsourcing.application.simple import SimpleApplication
from eventsourcing.application.snapshotting import SnapshottingApplication
from eventsourcing.application.sqlalchemy import SQLAlchemyApplication
from eventsourcing.domain.model.events import DomainEvent, assert_event_handlers_empty
from eventsourcing.exceptions import ProgrammingError
from eventsourcing.tests.core_tests.test_aggregate_root import ExampleAggregateRoot
from eventsourcing.tests.sequenced_item_tests.test_django_record_manager import (
    DjangoTestCase,
)
from eventsourcing.utils.random import encoded_random_bytes


class TestSimpleApplication(TestCase):

    application_class = SimpleApplication
    infrastructure_class = SQLAlchemyApplication

    def test_simple_application_without_infrastructure(self):
        with self.application_class() as app:

            with self.assertRaises(ProgrammingError):
                app.datastore

            with self.assertRaises(ProgrammingError):
                app.repository

            with self.assertRaises(ProgrammingError):
                app.event_store

            with self.assertRaises(ProgrammingError):
                app.notification_log

            with self.assertRaises(ProgrammingError):
                app.persistence_policy

    def test_application_with_infrastructure(self):
        with self.construct_concrete_application() as app:

            # Start with a new table.
            app.drop_table()
            app.drop_table()
            app.setup_table()
            app.setup_table()

            # Check the notifications.
            reader = NotificationLogReader(app.notification_log)
            old_notifications = reader.list_notifications()
            len_old = len(old_notifications)

            # Check the application's persistence policy,
            # repository, and event store, are working.
            aggregate = ExampleAggregateRoot.__create__()
            aggregate.__save__()
            self.assertTrue(aggregate.id in app.repository)

            # Check the notifications.
            reader = NotificationLogReader(app.notification_log)
            notifications = reader.list_notifications()
            self.assertEqual(1 + len_old, len(notifications))
            topic = "eventsourcing.tests.core_tests.test_aggregate_root#ExampleAggregateRoot.Created"
            self.assertEqual(topic, notifications[len_old]["topic"])

            app.drop_table()

    def construct_concrete_application(self):
        return self.application_class.mixin(self.infrastructure_class)(
            cipher_key=encoded_random_bytes(16), persist_event_type=DomainEvent
        )

    def tearDown(self):
        # Check the close() method leaves everything unsubscribed.
        assert_event_handlers_empty()


class TestPopoApplication(TestSimpleApplication):
    infrastructure_class = PopoApplication


class TestDjangoApplication(DjangoTestCase, TestSimpleApplication):
    infrastructure_class = DjangoApplication


class TestAxonApplication(TestSimpleApplication):
    infrastructure_class = AxonApplication


class TestSnapshottingApplication(TestSimpleApplication):
    infrastructure_class = SnapshottingApplication.mixin(SQLAlchemyApplication)


class TestSnapshottingAxonApplication(TestSimpleApplication):
    infrastructure_class = SnapshottingApplication.mixin(AxonApplication)



from eventsourcing.domain.model.collection import (
    AbstractCollectionRepository,
    Collection,
)
from eventsourcing.infrastructure.eventsourcedrepository import EventSourcedRepository


class CollectionRepository(EventSourcedRepository, AbstractCollectionRepository):
    """
    Event sourced repository for the Collection domain model entity.
    """



from typing import Any, Dict, Optional
from uuid import UUID

from eventsourcing.domain.model.events import (
    AbstractSnapshot,
    EventWithOriginatorID,
    EventWithOriginatorVersion,
    EventWithTimestamp,
)
from eventsourcing.utils.topic import reconstruct_object, resolve_topic
from eventsourcing.whitehead import TEntity


class Snapshot(
    EventWithTimestamp,
    EventWithOriginatorVersion,
    EventWithOriginatorID,
    AbstractSnapshot,
):
    def __init__(
        self,
        originator_id: UUID,
        originator_version: int,
        topic: str,
        state: Optional[Dict],
    ):
        super(Snapshot, self).__init__(
            originator_id=originator_id,
            originator_version=originator_version,
            topic=topic,
            state=state,
        )

    @property
    def topic(self) -> str:
        """
        Path to the class of the snapshotted entity.
        """
        return self.__dict__["topic"]

    @property
    def state(self) -> Dict[str, Any]:
        """
        State of the snapshotted entity.
        """
        return self.__dict__["state"]

    def __mutate__(self, obj: Optional[TEntity]) -> Optional[TEntity]:
        if self.state is not None:
            entity_class = resolve_topic(self.topic)
            return reconstruct_object(entity_class, self.state)



from eventsourcing.tests.datastore_tests import test_sqlalchemy
from eventsourcing.tests.sequenced_item_tests import base


class SQLAlchemyRecordManagerTestCase(
    test_sqlalchemy.SQLAlchemyDatastoreTestCase, base.WithRecordManagers
):
    """
    Base class for test cases that need record manager with SQLAlchemy.
    """


class TestSQLAlchemyRecordManagerWithIntegerSequences(
    SQLAlchemyRecordManagerTestCase, base.IntegerSequencedRecordTestCase
):
    """
    Test case for integer sequenced record manager with SQLAlchemy.
    """


class TestSQLAlchemyRecordManagerWithTimestampSequences(
    SQLAlchemyRecordManagerTestCase, base.TimestampSequencedItemTestCase
):
    """
    Test case for timestamp sequenced record manager with SQLAlchemy.
    """


class TestSimpleIteratorWithSQLAlchemy(
    SQLAlchemyRecordManagerTestCase, base.SequencedItemIteratorTestCase
):
    """
    Test case for simple iterator in record manager with SQLAlchemy.
    """


class TestThreadedIteratorWithSQLAlchemy(
    SQLAlchemyRecordManagerTestCase, base.ThreadedSequencedItemIteratorTestCase
):
    """
    Test case for threaded iterator in record manager with SQLAlchemy.
    """

    use_named_temporary_file = True



import logging
import os
import sys
import time
import unittest
from time import sleep
from unittest import skipIf
from uuid import UUID

import ray

from eventsourcing.application.popo import PopoApplication
from eventsourcing.application.sqlalchemy import SQLAlchemyApplication
from eventsourcing.domain.model.events import (
    assert_event_handlers_empty,
    clear_event_handlers,
)
from eventsourcing.system.definition import System
from eventsourcing.system.ray import RayProcess, RayRunner
from eventsourcing.system.rayhelpers import RayPrompt
from eventsourcing.tests.system_test_fixtures import (
    Orders,
    Payments,
    Reservations,
    set_db_uri,
)

logger = logging.getLogger("")
logger.setLevel(logging.ERROR)
ch = logging.StreamHandler()
ch.setLevel(logging.ERROR)
logger.addHandler(ch)


@skipIf(
    sys.version_info[:2] == (3, 6), "RayRunner not working with Python36 (pickle issue)"
)
# Seems to be doing something with a prompt, and type checking from annotation.
# It works in python3.7...
#
# Traceback (most recent call last):
#   File "...tests/test_ray_runner.py", line 61, in test_ray_runner
#     self.check_actors()
#   File "...tests/test_ray_runner.py", line 74, in check_actors
#     with runner:
#   File "...system/definition.py", line 236, in __enter__
#     self.start()
#   File "...system/ray.py", line 89, in start
#     setup_tables=self.setup_tables,
#   File "...ray/actor.py", line 326, in remote
#     return self._remote(args=args, kwargs=kwargs)
#   File "...ray/actor.py", line 479, in _remote
#     meta.modified_class, meta.actor_method_names)
#   File "...ray/function_manager.py", line 578, in export_actor_class
#     "class": pickle.dumps(Class),
#   File "...ray/cloudpickle/cloudpickle_fast.py", line 68, in dumps
#     cp.dump(obj)
#   File "...ray/cloudpickle/cloudpickle_fast.py", line 557, in dump
#     return Pickler.dump(self, obj)
# _pickle.PicklingError: Can't pickle typing.Union[
# eventsourcing.application.process.Prompt, NoneType]: it's not the same object as
# typing.Union


class TestRayRunner(unittest.TestCase):
    infrastructure_class = SQLAlchemyApplication
    # infrastructure_class = PopoApplication

    def setUp(self):
        # Define system.
        self.system = System(
            Orders | Reservations | Orders | Payments | Orders,
            infrastructure_class=self.infrastructure_class,
        )
        # Set environment.
        set_db_uri()

    def tearDown(self):
        # Unset environment.
        try:
            del os.environ["DB_URI"]
        except KeyError:
            pass

        try:
            assert_event_handlers_empty()
        finally:
            clear_event_handlers()

    def test_ray_runner(self):
        self.check_actors(2, 60)

    def check_actors(self, num_pipelines=1, num_orders_per_pipeline=10):

        pipeline_ids = list(range(num_pipelines))

        db_uri = os.environ.get("DB_URI", None)
        runner = RayRunner(
            self.system, pipeline_ids=pipeline_ids, setup_tables=True, db_uri=db_uri
        )

        num_orders = num_pipelines * num_orders_per_pipeline

        with runner:

            sleep(0.2)

            # Create some new orders.
            order_ids = []
            for i in range(num_orders):

                pipeline_id = i % len(pipeline_ids)
                order_id = runner.get(Orders, pipeline_id).create_new_order()
                order_ids.append(order_id)
                sleep(0.001)

            # Wait for orders to be reserved and paid.

            retries = 20 + 10 * num_orders
            for i, order_id in enumerate(order_ids):

                pipeline_id = i % len(pipeline_ids)

                orders = runner.get(Orders, pipeline_id)
                while not orders.is_order_reserved(order_id):
                    time.sleep(0.5)
                    retries -= 1
                    assert retries, "Failed set order.is_reserved {} ({})".format(
                        order_id, i
                    )

                while not orders.is_order_paid(order_id):
                    time.sleep(0.5)
                    retries -= 1
                    assert retries, "Failed set order.is_paid ({})".format(i)

                # print(i, "Completed order", order_id)

            if not order_ids:
                return

            # Calculate timings from event timestamps.
            orders = []
            for i, order_id in enumerate(order_ids):
                pipeline_id = i % len(pipeline_ids)
                orders.append(runner.get(Orders, pipeline_id).get_order(order_id))
            first_timestamp = min([o.__created_on__ for o in orders])
            last_timestamp = max([o.__last_modified__ for o in orders])
            duration = last_timestamp - first_timestamp
            rate = len(order_ids) / float(duration)
            period = 1 / rate
            print(
                "Orders system processed {} orders in {:.3f}s at rate of {:.1f} "
                "orders/s, {:.3f}s each".format(len(order_ids), duration, rate, period)
            )

            # Print min, average, max duration.
            durations = [o.__last_modified__ - o.__created_on__ for o in orders]
            print("Min order processing time: {:.3f}s".format(min(durations)))
            print(
                "Mean order processing time: {:.3f}s".format(
                    sum(durations) / len(durations)
                )
            )
            print("Max order processing time: {:.3f}s".format(max(durations)))

            sleep(0)


class TestRayProcess(unittest.TestCase):
    infrastructure_class = SQLAlchemyApplication

    def test_ray_process(self):
        # Create process with Orders application.
        ray_orders_process = RayProcess.remote(
            application_process_class=Orders,
            infrastructure_class=self.infrastructure_class,
            setup_tables=True,
        )

        # Initialise the ray process.
        ray.get(ray_orders_process.init.remote({}, {}))

        # Create a new order, within the ray process.
        order_id = ray.get(ray_orders_process.call.remote("create_new_order"))

        # Check a UUID is returned.
        self.assertIsInstance(order_id, UUID)

        # Get range of notifications.
        notifications = ray.get(ray_orders_process.get_notifications.remote(1, 1000))
        self.assertIsInstance(notifications, list)
        self.assertEqual(len(notifications), 1)
        self.assertEqual(notifications[0]["id"], 1)

        # Check the range is working.
        notifications = ray.get(ray_orders_process.get_notifications.remote(2, 2))
        self.assertIsInstance(notifications, list)
        self.assertEqual(len(notifications), 0, notifications)

        # Create process with Reservations application.
        ray_reservations_process = RayProcess.remote(
            application_process_class=Reservations,
            infrastructure_class=self.infrastructure_class,
            setup_tables=True,
        )
        # Make the reservations follow the orders.
        ray.get(
            ray_reservations_process.init.remote({"orders": ray_orders_process}, {})
        )

        # Get range of notifications.
        notifications = ray.get(
            ray_reservations_process.get_notifications.remote(1, 1000)
        )
        self.assertIsInstance(notifications, list)
        self.assertEqual(len(notifications), 0)

        # Prompt the process.
        prompt = RayPrompt("orders", 0)
        rayid = ray_reservations_process.prompt.remote(prompt)
        ray.get(rayid)

        # Check a reservation was created.
        retries = 10
        while retries:
            sleep(0.1)
            # Get range of notifications.
            notifications = ray.get(
                ray_reservations_process.get_notifications.remote(1, 1000)
            )
            self.assertIsInstance(notifications, list)

            try:
                self.assertEqual(len(notifications), 1)
            except AssertionError:
                if retries:
                    retries -= 1
                    print("Retrying...", retries)
                else:
                    raise
            else:
                break

        # Add reservations as downstream process of orders.
        ray_orders_process.add_downstream_process.remote(
            "reservations", ray_reservations_process
        )

        # Create a new order, within the ray process.
        order_id = ray.get(ray_orders_process.call.remote("create_new_order"))

        # Get range of notifications.
        notifications = ray.get(ray_orders_process.get_notifications.remote(1, 1000))
        self.assertIsInstance(notifications, list)
        self.assertEqual(len(notifications), 2)

        # Check another reservation was created.
        retries = 10
        while True:
            sleep(0.1)
            # Get a section of the notification log.
            # Get range of notifications.
            notifications = ray.get(
                ray_reservations_process.get_notifications.remote(1, 1000)
            )
            self.assertIsInstance(notifications, list)

            try:
                self.assertEqual(len(notifications), 2)
            except AssertionError:
                if retries:
                    retries -= 1
                    print("Retrying...", retries)
                else:
                    raise
            else:
                break

        # Todo: More of this...

        # reader = NotificationLogReader(
        #     notification_log=RayNotificationLog(
        #         upstream_process=ray_process
        #     ),
        # )
        # notifications = list(reader.read())
        # self.assertEqual(len(notifications), 1)
        # created_event_notification = notifications[0]
        # self.assertEqual(created_event_notification['id'], 1)
        # self.assertEqual(created_event_notification['originator_id'], order_id)
        # self.assertEqual(created_event_notification['originator_version'], 0)



from abc import ABC, abstractmethod
from json import JSONDecodeError
from typing import Any, Dict, Generic, NamedTuple, Optional, Tuple, Type

from eventsourcing.infrastructure.sequenceditem import (
    SequencedItem,
    SequencedItemFieldNames,
)
from eventsourcing.utils.cipher.aes import AESCipher
from eventsourcing.utils.topic import get_topic, reconstruct_object, resolve_topic
from eventsourcing.utils.transcoding import ObjectJSONDecoder, ObjectJSONEncoder
from eventsourcing.whitehead import TEvent


class AbstractSequencedItemMapper(Generic[TEvent], ABC):
    def __init__(self, **kwargs: Any):
        """
        Initialises mapper.
        """

    @abstractmethod
    def item_from_event(self, domain_event: TEvent) -> NamedTuple:
        """
        Constructs and returns a sequenced item for given domain event.
        """

    @abstractmethod
    def event_from_item(self, sequenced_item: NamedTuple) -> TEvent:
        """
        Constructs and returns a domain event for given sequenced item.
        """

    @abstractmethod
    def json_dumps(self, o: object) -> bytes:
        """
        Encodes given object as JSON.
        """

    @abstractmethod
    def json_loads(self, s: str) -> object:
        """
        Decodes given JSON as object.
        """

    @abstractmethod
    def event_from_topic_and_state(self, topic: str, state: bytes) -> TEvent:
        """
        Resolves topic to an event class, decodes state, and constructs an event.
        """


class SequencedItemMapper(AbstractSequencedItemMapper[TEvent]):
    """
    Uses JSON to transcode domain events.
    """

    def __init__(
        self,
        sequenced_item_class: Optional[Type[NamedTuple]] = None,
        sequence_id_attr_name: Optional[str] = None,
        position_attr_name: Optional[str] = None,
        json_encoder_class: Optional[Type[ObjectJSONEncoder]] = None,
        sort_keys: bool = False,
        json_decoder_class: Optional[Type[ObjectJSONDecoder]] = None,
        cipher: Optional[AESCipher] = None,
        compressor: Any = None,
        other_attr_names: Tuple[str, ...] = (),
    ):
        if sequenced_item_class is not None:
            self.sequenced_item_class = sequenced_item_class
        else:
            self.sequenced_item_class = SequencedItem  # type: ignore
        self.json_encoder_class = json_encoder_class or ObjectJSONEncoder
        self.json_encoder = self.json_encoder_class(sort_keys=sort_keys)
        self.json_decoder_class = json_decoder_class or ObjectJSONDecoder
        self.json_decoder = self.json_decoder_class()
        self.cipher = cipher
        self.compressor = compressor
        self.field_names = SequencedItemFieldNames(self.sequenced_item_class)
        self.sequence_id_attr_name = (
            sequence_id_attr_name or self.field_names.sequence_id
        )
        self.position_attr_name = position_attr_name or self.field_names.position
        self.other_attr_names = other_attr_names or self.field_names.other_names

    def item_from_event(self, domain_event: TEvent) -> NamedTuple:
        """
        Constructs a sequenced item from a domain event.
        """
        item_args = self.construct_item_args(domain_event)
        return self.construct_sequenced_item(item_args)

    def construct_item_args(self, domain_event: TEvent) -> Tuple:
        """
        Constructs attributes of a sequenced item from the given domain event.
        """
        # Get the sequence ID.
        sequence_id = domain_event.__dict__[self.sequence_id_attr_name]

        # Get the position in the sequence.
        position = getattr(domain_event, self.position_attr_name, None)

        # Get topic and data.
        topic, state = self.get_item_topic_and_state(
            domain_event.__class__, domain_event.__dict__
        )

        # Get the 'other' args.
        # - these are meant to be derivative of the other attributes,
        #   to populate database fields, and shouldn't affect the hash.
        other_args = tuple(
            (getattr(domain_event, name) for name in self.other_attr_names)
        )

        return (sequence_id, position, topic, state) + other_args

    def get_item_topic_and_state(
        self, domain_event_class: type, event_attrs: Dict[str, Any]
    ) -> Tuple[str, bytes]:
        # Get the topic from the event attrs, otherwise from the class.
        topic = get_topic(domain_event_class)

        # Serialise the event attributes.
        statebytes = self.json_dumps(event_attrs)

        # Compress plaintext bytes.
        if self.compressor:
            # Zlib reduces length by about 25% to 50%.
            statebytes = self.compressor.compress(statebytes)

        # Encrypt serialised state.
        if self.cipher:
            # Increases length by about 10%.
            statebytes = self.cipher.encrypt(statebytes)

        return topic, statebytes

    def json_dumps(self, o: object) -> bytes:
        return self.json_encoder.encode(o)

    def construct_sequenced_item(self, item_args: Tuple) -> NamedTuple:
        return self.sequenced_item_class(*item_args)

    def event_from_item(self, sequenced_item: NamedTuple) -> TEvent:
        """
        Reconstructs domain event from stored event topic and
        event attrs. Used in the event store when getting domain events.
        """
        assert isinstance(sequenced_item, self.sequenced_item_class), (
            self.sequenced_item_class,
            type(sequenced_item),
        )

        # Get the topic and state.
        topic = getattr(sequenced_item, self.field_names.topic)
        state = getattr(sequenced_item, self.field_names.state)

        return self.event_from_topic_and_state(topic, state)

    def event_from_topic_and_state(self, topic: str, state: bytes) -> TEvent:
        domain_event_class, event_attrs = self.get_event_class_and_attrs(topic, state)

        # Reconstruct domain event object.
        return reconstruct_object(domain_event_class, event_attrs)

    def get_event_class_and_attrs(
        self, topic: str, state: bytes
    ) -> Tuple[Type[TEvent], Dict]:
        # Resolve topic to event class.
        domain_event_class: Type[TEvent] = resolve_topic(topic)

        # Decrypt and decompress state.
        if self.cipher:
            state = self.cipher.decrypt(state)

        # Decompress plaintext bytes.
        if self.compressor:
            state = self.compressor.decompress(state)

        # Decode unicode bytes.
        statestr = state.decode("utf8")

        # Deserialize JSON.
        event_attrs: Dict = self.json_loads(statestr)

        # Return instance class and attribute values.
        return domain_event_class, event_attrs

    def json_loads(self, s: str) -> Dict:
        try:
            return self.json_decoder.decode(s)
        except JSONDecodeError:
            raise ValueError("Couldn't load JSON string: {}".format(s))



from math import ceil
from threading import Thread
from uuid import uuid4

from eventsourcing.application.notificationlog import (
    BigArrayNotificationLog,
    NotificationLogReader,
    RecordManagerNotificationLog,
)
from eventsourcing.domain.model.events import DomainEvent
from eventsourcing.infrastructure.repositories.array import BigArrayRepository
from eventsourcing.interface.notificationlog import (
    NotificationLogView,
    RemoteNotificationLog,
)
from eventsourcing.tests.sequenced_item_tests.base import WithEventPersistence
from eventsourcing.tests.sequenced_item_tests.test_cassandra_record_manager import (
    WithCassandraRecordManagers,
)
from eventsourcing.tests.sequenced_item_tests.test_django_record_manager import (
    DjangoTestCase,
)
from eventsourcing.tests.sequenced_item_tests.test_sqlalchemy_record_manager import (
    SQLAlchemyRecordManagerTestCase,
)
from eventsourcing.utils.topic import get_topic
from eventsourcing.utils.transcoding import ObjectJSONEncoder


class NotificationLogTestCase(SQLAlchemyRecordManagerTestCase, WithEventPersistence):
    def assert_section(
        self,
        repo,
        requested_id,
        expected_id,
        expected_len_items,
        expected_previous_id,
        expected_next_id,
    ):
        section = repo[requested_id]
        self.assertEqual(expected_len_items, len(list(section.items)))
        self.assertEqual(expected_id, section.section_id)
        self.assertEqual(expected_previous_id, section.previous_id)
        self.assertEqual(expected_next_id, section.next_id)

    def append_notifications(self, *range_args):
        for i in range(*range_args):
            item = ("item{}".format(i + 1)).encode("utf8")
            self.append_notification(item)

    def create_notification_log(self, section_size):
        return RecordManagerNotificationLog(self.entity_record_manager, section_size)

    def construct_entity_record_manager(self):
        return self.factory.construct_integer_sequenced_record_manager()

    def append_notification(self, item):
        sequenced_item = self.entity_record_manager.sequenced_item_class(
            uuid4(), 0, get_topic(DomainEvent), item
        )
        self.entity_record_manager.record_item(sequenced_item)


class TestNotificationLog(NotificationLogTestCase):
    def test(self):
        # Build notification log.
        section_size = 5
        notification_log = self.create_notification_log(section_size=section_size)

        # Check the sections.
        section = notification_log["current"]
        self.assertEqual("1,5", section.section_id)
        self.assertEqual(0, len(list(section.items)))
        self.assertIsNone(section.previous_id)
        self.assertIsNone(section.previous_id)

        # Append notifications.
        self.append_notifications(13)

        # Check the sections.
        self.assert_section(notification_log, "current", "11,15", 3, "6,10", None)
        self.assert_section(notification_log, "1,5", "1,5", section_size, None, "6,10")
        self.assert_section(
            notification_log, "6,10", "6,10", section_size, "1,5", "11,15"
        )
        self.assert_section(notification_log, "11,15", "11,15", 3, "6,10", None)
        self.assert_section(notification_log, "16,20", "16,20", 0, "11,15", None)
        self.assert_section(notification_log, "21,25", "21,25", 0, "16,20", None)

        # Add some more notification.
        self.append_notifications(13, 24)

        # Check the notification log has been extended.
        self.assertEqual(len(list(notification_log["11,15"].items)), section_size)
        self.assertEqual(len(list(notification_log["16,20"].items)), section_size)
        self.assertEqual(len(list(notification_log["21,25"].items)), 4)
        self.assertEqual(len(list(notification_log["26,30"].items)), 0)

        # Check an section ID that can't be split by ',' results in a ValueError.
        with self.assertRaises(ValueError):
            _ = notification_log["invalid"]


class TestBigArrayNotificationLog(TestNotificationLog):
    def test(self):
        self.section_size = 5
        super(TestBigArrayNotificationLog, self).test()

        # Check array size must be divisible by section size.
        with self.assertRaises(ValueError):
            BigArrayNotificationLog(self.big_array, section_size=6)

        # # Check the section ID must match the section size.
        # notification_log = BigArrayNotificationLog(self.big_array, self.section_size)
        # with self.assertRaises(ValueError):
        #     _ = notification_log['1,2']

        # # Check the section ID must be aligned to the array size.
        # with self.assertRaises(ValueError):
        #     _ = notification_log['2,6']

    def create_notification_log(self, section_size):
        self.big_array = self.create_big_array()
        return BigArrayNotificationLog(self.big_array, section_size=section_size)

    def create_big_array(self, big_array_id=None):
        big_array_id = uuid4() if big_array_id is None else big_array_id
        big_array_repo = BigArrayRepository(event_store=self.entity_event_store)
        big_array = big_array_repo[big_array_id]
        return big_array

    def append_notification(self, item):
        self.big_array.append(item)


class TestNotificationLogReader(NotificationLogTestCase):
    def test(self):
        # Build notification log.
        section_size = 5
        notification_log = self.create_notification_log(section_size=section_size)

        # Append 13 notifications.
        self.append_notifications(13)

        # Construct notification log reader.
        reader = NotificationLogReader(notification_log)

        # Check position.
        self.assertEqual(reader.position, 0)

        # Read all notifications.
        all_notifications = list(reader)
        self.assertEqual(13, len(all_notifications))

        # Check position.
        self.assertEqual(reader.position, 13)

        # Add some more items to the log.
        self.append_notifications(13, 21)

        # Read subsequent notifications.
        subsequent_notifications_notifications = list(reader)
        self.assertEqual(len(subsequent_notifications_notifications), 8)

        # Check position.
        self.assertEqual(reader.position, 21)

        subsequent_notifications_notifications = list(reader)
        self.assertEqual(len(subsequent_notifications_notifications), 0)

        # Set position.
        reader.seek(13)
        subsequent_notifications_notifications = list(reader)
        self.assertEqual(len(subsequent_notifications_notifications), 8)

        # # Read items after a particular position.
        self.assertEqual(len(list(reader[0:])), 21)
        self.assertEqual(len(list(reader[1:])), 20)
        self.assertEqual(len(list(reader[2:])), 19)
        self.assertEqual(len(list(reader[3:])), 18)
        self.assertEqual(len(list(reader[13:])), 8)
        self.assertEqual(len(list(reader[18:])), 3)
        self.assertEqual(len(list(reader[19:])), 2)
        self.assertEqual(len(list(reader[20:])), 1)
        self.assertEqual(len(list(reader[21:])), 0)

        # Check last item numbers less than 1 cause a value errors.
        with self.assertRaises(ValueError):
            reader.position = -1
            list(reader)

        with self.assertRaises(ValueError):
            list(reader.seek(-1))

        # Resume from a saved position.
        saved_position = 5
        advance_by = 3
        reader.seek(saved_position)
        self.assertEqual(reader.position, saved_position)
        reader.list_notifications(advance_by=advance_by)
        self.assertEqual(reader.position, saved_position + advance_by)

        # Read items between particular positions.
        # - check stops at end of slice, and position tracks ok
        self.assertEqual(reader[0]["id"], 1)
        self.assertEqual(reader.position, 1)

        self.assertEqual(next(reader)["id"], 2)
        self.assertEqual(reader.position, 2)

        reader.seek(5)
        self.assertEqual(next(reader)["id"], 6)
        self.assertEqual(reader.position, 6)

        reader.seek(0)
        list(reader)
        self.assertEqual(reader.position, 21)

        self.assertEqual(len(list(reader[0:1])), 1)
        self.assertEqual(reader.position, 1)

        self.assertEqual(len(list(reader[1:3])), 2)
        self.assertEqual(reader.position, 3)

        self.assertEqual(len(list(reader[2:5])), 3)
        self.assertEqual(reader.position, 5)

        self.assertEqual(len(list(reader[3:7])), 4)
        self.assertEqual(reader.position, 7)

        self.assertEqual(len(list(reader[13:20])), 7)
        self.assertEqual(reader.position, 20)

        self.assertEqual(len(list(reader[18:20])), 2)
        self.assertEqual(reader.position, 20)

        self.assertEqual(len(list(reader[19:20])), 1)
        self.assertEqual(reader.position, 20)

        self.assertEqual(len(list(reader[20:20])), 0)
        self.assertEqual(reader.position, 20)

        self.assertEqual(len(list(reader[21:20])), 0)
        self.assertEqual(reader.position, 21)

        with self.assertRaises(StopIteration):
            next(reader)


class TestRemoteNotificationLog(NotificationLogTestCase):
    use_named_temporary_file = True

    def test_remote_notification_log(self):
        num_notifications = 42

        section_size = 5

        # Build a notification log (fixture).
        self.append_notifications(num_notifications)

        # Start a simple server.
        from wsgiref.util import setup_testing_defaults
        from wsgiref.simple_server import make_server

        port = 8080
        base_url = "http://127.0.0.1:{}/notifications/".format(port)

        def simple_app(environ, start_response):
            """Simple WSGI application."""
            setup_testing_defaults(environ)

            # Identify log and section from request.
            path_info = environ["PATH_INFO"]
            try:
                section_id = path_info.strip("/").split("/")[-1]
            except ValueError:
                # Start response.
                status = "404 Not Found"
                headers = [("Content-type", "text/plain; charset=utf-8")]
                start_response(status, headers)
                return []

            # Select the notification log.
            notification_log = self.create_notification_log(section_size)

            # Get serialized section.
            json_encoder = ObjectJSONEncoder()
            view = NotificationLogView(notification_log, json_encoder)

            resource = view.present_resource(section_id)
            # Todo: Maybe redirect if the section ID is a mismatch, so
            # the URL is good for cacheing.

            # Start response.
            status = "200 OK"
            headers = [("Content-type", "text/plain; charset=utf-8")]
            start_response(status, headers)

            # Return a list of lines.
            return [resource]

        httpd = make_server("", port, simple_app)
        print("Serving on port {}...".format(port))
        thread = Thread(target=httpd.serve_forever)
        thread.setDaemon(True)
        thread.start()
        try:
            # Use reader with client to read all items in remote feed after item 5.
            notification_log = RemoteNotificationLog(base_url)

            # Just before we start, test the deserialise_section_size exceptions.
            notification_log.deserialize_section_size("1")
            with self.assertRaises(ValueError):
                notification_log.deserialize_section_size('"1')
            with self.assertRaises(TypeError):
                notification_log.deserialize_section_size('"1"')

            # Get all the items.
            notification_log_reader = NotificationLogReader(
                notification_log=notification_log
            )
            items_from_start = notification_log_reader.list_notifications()

            # Check we got all the items.
            self.assertEqual(len(items_from_start), num_notifications)
            self.assertEqual(items_from_start[0]["id"], 1)
            self.assertEqual(items_from_start[0]["state"], b"item1")
            self.assertEqual(
                items_from_start[0]["topic"],
                "eventsourcing.domain.model.events#DomainEvent",
            )
            expected_section_count = ceil(num_notifications / float(section_size))
            self.assertEqual(
                notification_log_reader.section_count, expected_section_count
            )

            # Get all the items from item 5.
            items_from_5 = list(notification_log_reader[section_size - 1 :])

            # Check we got everything after item 5.
            self.assertEqual(len(items_from_5), num_notifications - section_size + 1)
            self.assertEqual(items_from_5[0]["id"], section_size)
            self.assertEqual(
                items_from_5[0]["topic"],
                "eventsourcing.domain.model.events#DomainEvent",
            )
            self.assertEqual(
                items_from_5[0]["state"], "item{}".format(section_size).encode("utf8")
            )
            expected_section_count = ceil(num_notifications / float(section_size))
            self.assertEqual(
                notification_log_reader.section_count, expected_section_count
            )

            # Check ValueError is raised for deserialization errors.
            with self.assertRaises(ValueError):
                notification_log.deserialize_section("invalid json")

        finally:
            httpd.shutdown()
            thread.join()
            httpd.server_close()


class TestNotificationLogWithDjango(DjangoTestCase, TestNotificationLog):
    pass


class TestBigArrayNotificationLogWithDjango(
    DjangoTestCase, TestBigArrayNotificationLog
):
    pass


class TestBigArrayNotificationLogWithCassandra(
    WithCassandraRecordManagers, TestBigArrayNotificationLog
):
    pass



from decimal import Decimal
from unittest import TestCase
from uuid import uuid4

from sqlalchemy import Column, Text
from sqlalchemy_utils import UUIDType

from eventsourcing.application.command import CommandProcess
from eventsourcing.application.process import (
    ProcessApplication,
    ProcessApplicationWithSnapshotting,
    WrappedRepository,
)
from eventsourcing.application.simple import PromptToPull
from eventsourcing.application.sqlalchemy import SQLAlchemyApplication
from eventsourcing.domain.model.aggregate import AggregateRoot, BaseAggregateRoot
from eventsourcing.domain.model.command import Command
from eventsourcing.domain.model.events import (
    EventHandlersNotEmptyError,
    assert_event_handlers_empty,
    clear_event_handlers,
    subscribe,
    unsubscribe,
)
from eventsourcing.domain.model.snapshot import Snapshot
from eventsourcing.exceptions import (
    CausalDependencyFailed,
    ProgrammingError,
    PromptFailed,
)
from eventsourcing.infrastructure.sqlalchemy.records import Base
from eventsourcing.utils.topic import resolve_topic
from eventsourcing.utils.transcoding import ObjectJSONDecoder
from eventsourcing.whitehead import TEntity


class TestProcessApplication(TestCase):
    infrastructure_class = SQLAlchemyApplication

    def test_process_with_example_policy(self):
        # Construct example process.
        process_class = ProcessApplication.mixin(self.infrastructure_class)
        with process_class(
            name="test",
            policy=example_policy,
            persist_event_type=ExampleAggregate.Event,
            setup_table=True,
        ) as process:
            # Make the process follow itself.
            process.follow("test", process.notification_log)

            # Create an aggregate.
            aggregate = ExampleAggregate.__create__()
            aggregate.__save__()

            # Run the process.
            process.run()

            # Check the aggregate has been "moved on".
            self.assertTrue(process.repository[aggregate.id].is_moved_on)

            # Check the __contains__ method of the repo wrapper.
            self.assertIn(aggregate.id, WrappedRepository(process.repository))
            self.assertNotIn(uuid4(), WrappedRepository(process.repository))

            # Check the repository wrapper tracks causal dependencies.
            repository = WrappedRepository(process.repository)
            aggregate = repository[aggregate.id]
            causal_dependencies = repository.causal_dependencies
            self.assertEqual(len(causal_dependencies), 1)
            self.assertEqual((aggregate.id, 1), causal_dependencies[0])

            # Check events from more than one aggregate are stored.
            self.assertIn(aggregate.second_id, process.repository)

    def test_process_application_with_snapshotting(self):
        # Construct example process.
        with ProcessApplicationWithSnapshotting.mixin(self.infrastructure_class)(
            name="test",
            policy=example_policy,
            persist_event_type=ExampleAggregate.Event,
            setup_table=True,
            snapshot_period=2,
        ) as process:
            # Make the process follow itself.
            process.follow("test", process.notification_log)

            # Create an aggregate.
            aggregate = ExampleAggregate.__create__()

            # Check there isn't a snapshot.
            self.assertIsNone(process.snapshot_strategy.get_snapshot(aggregate.id))

            # Should "move on" by the process following itself.
            aggregate.__save__()
            process.run()

            aggregate = process.repository[aggregate.id]
            self.assertEqual(1, aggregate.__version__)

            # Check there is a snapshot.
            snapshot = process.snapshot_strategy.get_snapshot(aggregate.id)
            self.assertIsInstance(snapshot, Snapshot)
            self.assertEqual(snapshot.originator_version, 1)

            snapshot_v0 = process.snapshot_strategy.get_snapshot(
                aggregate.id, lt=snapshot.originator_version
            )
            self.assertIsNone(snapshot_v0, Snapshot)

    def test_causal_dependencies(self):
        # Try to process an event that has unresolved causal dependencies.
        pipeline_id1 = 0
        pipeline_id2 = 1

        # Create two events, one has causal dependency on the other.
        process_class = ProcessApplication.mixin(self.infrastructure_class)
        core1 = process_class(
            name="core",
            # persist_event_type=ExampleAggregate.Created,
            persist_event_type=BaseAggregateRoot.Event,
            setup_table=True,
            pipeline_id=pipeline_id1,
        )
        core1.use_causal_dependencies = True

        kwargs = {}
        if self.infrastructure_class.is_constructed_with_session:
            # Needed for SQLAlchemy only.
            kwargs["session"] = core1.session

        core2 = process_class(
            name="core", pipeline_id=pipeline_id2, policy=example_policy, **kwargs
        )
        core2.use_causal_dependencies = True

        # First event in pipeline 1.
        aggregate = ExampleAggregate.__create__()
        aggregate.__save__()

        # Second event in pipeline 2.
        # - it's important this is done in a policy so the causal dependencies are
        # identified
        core2.follow("core", core1.notification_log)
        core2.run()

        # Check the aggregate exists.
        self.assertTrue(aggregate.id in core1.repository)

        # Check the aggregate has been "moved on".
        aggregate = core1.repository[aggregate.id]
        self.assertTrue(aggregate.is_moved_on)
        self.assertTrue(aggregate.second_id)
        self.assertIn(aggregate.second_id, core1.repository)

        # Check the events have different pipeline IDs.
        aggregate_records = core1.event_store.record_manager.get_records(aggregate.id)
        second_entity_records = core1.event_store.record_manager.get_records(
            aggregate.second_id
        )

        self.assertEqual(2, len(aggregate_records))
        self.assertEqual(1, len(second_entity_records))

        self.assertEqual(pipeline_id1, aggregate_records[0].pipeline_id)
        self.assertEqual(pipeline_id2, aggregate_records[1].pipeline_id)
        self.assertEqual(pipeline_id2, second_entity_records[0].pipeline_id)

        # Check the causal dependencies have been constructed.
        # - the first 'Created' event doesn't have an causal dependencies
        self.assertFalse(aggregate_records[0].causal_dependencies)

        # - the second 'Created' event depends on the Created event in another pipeline.
        expect = [{"notification_id": 1, "pipeline_id": pipeline_id1}]
        actual = ObjectJSONDecoder().decode(
            second_entity_records[0].causal_dependencies
        )

        self.assertEqual(expect, actual)

        # - the 'AttributeChanged' event depends on the second Created,
        # which is in the same pipeline, so expect no causal dependencies.
        self.assertFalse(aggregate_records[1].causal_dependencies)

        # Setup downstream process.
        downstream1 = process_class(
            name="downstream",
            pipeline_id=pipeline_id1,
            policy=event_logging_policy,
            **kwargs
        )
        downstream1.follow("core", core1.notification_log)
        downstream2 = process_class(
            name="downstream",
            pipeline_id=pipeline_id2,
            policy=event_logging_policy,
            **kwargs
        )
        downstream2.follow("core", core2.notification_log)

        # Try to process pipeline 2, should fail due to causal dependency.
        with self.assertRaises(CausalDependencyFailed):
            downstream2.run()

        self.assertEqual(
            0,
            len(
                list(downstream1.event_store.record_manager.get_notification_records())
            ),
        )
        self.assertEqual(
            0,
            len(
                list(downstream2.event_store.record_manager.get_notification_records())
            ),
        )

        # Try to process pipeline 1, should work.
        downstream1.run()

        self.assertEqual(
            1,
            len(
                list(downstream1.event_store.record_manager.get_notification_records())
            ),
        )
        self.assertEqual(
            0,
            len(
                list(downstream2.event_store.record_manager.get_notification_records())
            ),
        )

        # Try again to process pipeline 2, should work this time.
        downstream2.run()

        self.assertEqual(
            1,
            len(
                list(downstream1.event_store.record_manager.get_notification_records())
            ),
        )
        self.assertEqual(
            2,
            len(
                list(downstream2.event_store.record_manager.get_notification_records())
            ),
        )

        core1.close()
        core2.close()
        downstream1.close()
        downstream2.close()

    def test_projection_into_custom_orm_obj(self):

        self.define_projection_record_class()

        projection_id = uuid4()

        def projection_policy(repository: WrappedRepository, event):
            if isinstance(event, ExampleAggregate.Created):
                projection = self.projection_record_class(
                    projection_id=projection_id, state=str(event.timestamp)
                )
                repository.save_orm_obj(projection)
            elif isinstance(event, ExampleAggregate.MovedOn):
                projection = self.get_projection_record(
                    projection_record_class=self.projection_record_class,
                    projection_id=projection_id,
                    record_manager=repository.repository.event_store.record_manager,
                )
                assert projection is not None
                repository.delete_orm_obj(projection)

        # Construct example process.
        process_class = ProcessApplication.mixin(self.infrastructure_class)
        with process_class(
            name="test",
            policy=projection_policy,
            persist_event_type=ExampleAggregate.Event,
            setup_table=True,
        ) as process:
            assert isinstance(process, ProcessApplication)

            self.setup_projections_table(process)

            # Make the process follow itself.
            process.follow("test", process.notification_log)

            # Create an aggregate.
            aggregate = ExampleAggregate.__create__()
            aggregate.__save__()

            # Run the process.
            process.run()

            # Check the projection has been created.
            projection = self.get_projection_record(
                projection_record_class=self.projection_record_class,
                projection_id=projection_id,
                record_manager=process.event_store.record_manager,
            )
            self.assertIsInstance(projection, self.projection_record_class)
            self.assertEqual(Decimal(projection.state), aggregate.___created_on__)

            # Move aggregate on.
            aggregate.move_on()
            aggregate.__save__()

            # Run the process.
            process.run()

            # Check the projection has been deleted.
            projection = self.get_projection_record(
                projection_record_class=self.projection_record_class,
                projection_id=projection_id,
                record_manager=process.event_store.record_manager,
            )
            self.assertIsNone(projection)

    def test_cant_follow_self_and_apply_policy_to_generated_events(self):
        with self.assertRaises(ProgrammingError):
            # Construct example process.
            process_class = ProcessApplication.mixin(self.infrastructure_class)
            with process_class(
                name="test",
                policy=None,
                persist_event_type=ExampleAggregate.Event,
                setup_table=True,
                apply_policy_to_generated_events=True,
            ) as process:
                assert isinstance(process, ProcessApplication)

                self.assertTrue(process.apply_policy_to_generated_events)

                # Make the process follow itself.
                process.follow("test", process.notification_log)

    def test_clear_cache_after_exception_recording_event(self):
        # Construct example process.
        process_class = ProcessApplication.mixin(self.infrastructure_class)
        with process_class(
            name="test",
            policy=example_policy,
            persist_event_type=ExampleAggregate.Event,
            setup_table=True,
            use_cache=True,
        ) as process:
            # Check cache is being used.
            self.assertTrue(process.use_cache)

            # Make the process follow itself.
            process.follow("test", process.notification_log)

            # Create an aggregate.
            aggregate = ExampleAggregate.__create__()
            aggregate.__save__()

            # Put aggregate in the cache.
            process.repository._cache[aggregate.id] = aggregate

            # Damage the record manager.
            process.event_store.record_manager.write_records = None

            # Run the process.
            try:
                process.run()
            except Exception:
                pass
            else:
                self.fail("Exception not raised")

            # Check the aggregate is no longer in the cache.
            self.assertNotIn(aggregate.id, process.repository._cache)

    def test_policy_returns_sequence_of_new_aggregates(self):
        def policy(repository, event):
            first = AggregateRoot.__create__(originator_id=(uuid4()))
            second = AggregateRoot.__create__(originator_id=(uuid4()))
            return (first, second)

        # Construct example process.
        process_class = ProcessApplication.mixin(self.infrastructure_class)
        with process_class(
            name="test",
            policy=policy,
            persist_event_type=ExampleAggregate.Event,
            setup_table=True,
            use_cache=True,
        ) as process:
            # Make the process follow itself.
            process.follow("test", process.notification_log)

            # Create an aggregate.
            aggregate = ExampleAggregate.__create__()
            aggregate.__save__()

            # Run the process.
            process.run()

            # Check the two new aggregates are in the cache.
            self.assertEqual(len(process.repository._cache), 2)

            # Check there are three aggregates in the repository.
            ids = process.repository.event_store.record_manager.all_sequence_ids()
            self.assertEqual(len(list(ids)), 3)

    def define_projection_record_class(self):
        class ProjectionRecord(Base):
            __tablename__ = "projections"

            # Projection ID.
            projection_id = Column(UUIDType(), primary_key=True)

            # State of the projection (serialized dict, possibly encrypted).
            state = Column(Text())

        self.projection_record_class = ProjectionRecord

    def setup_projections_table(self, process):
        process.datastore.setup_table(self.projection_record_class)

    def get_projection_record(
        self, projection_record_class, projection_id, record_manager
    ):
        return record_manager.session.query(projection_record_class).get(projection_id)

    def test_handle_prompt_failed(self):
        process = ProcessApplication.mixin(self.infrastructure_class)(
            name="test",
            policy=example_policy,
            persist_event_type=ExampleAggregate.Event,
            setup_table=True,
        )

        def raise_exception(_):
            raise Exception()

        def raise_prompt_failed(_):
            raise PromptFailed()

        subscribe(raise_exception)
        try:
            with self.assertRaises(PromptFailed):
                process.publish_prompt_for_events()
        finally:
            unsubscribe(raise_exception)

        subscribe(raise_prompt_failed)
        try:
            with self.assertRaises(PromptFailed):
                process.publish_prompt_for_events()
        finally:
            unsubscribe(raise_prompt_failed)

        try:
            process.publish_prompt_for_events()
        finally:
            process.close()

    def tearDown(self):
        try:
            assert_event_handlers_empty()
        except EventHandlersNotEmptyError:
            clear_event_handlers()
            raise


class TestPromptToPull(TestCase):
    def test_repr(self):
        prompt1 = PromptToPull("process1", pipeline_id=1)
        self.assertEqual(
            repr(prompt1), "PromptToPull(process_name=process1, pipeline_id=1)"
        )

    def test_eq(self):
        prompt1 = PromptToPull("process1", pipeline_id=1)
        prompt2 = PromptToPull("process1", pipeline_id=1)
        prompt3 = PromptToPull("process2", pipeline_id=1)
        prompt4 = PromptToPull("process1", pipeline_id=2)

        self.assertEqual(prompt1, prompt1)
        self.assertEqual(prompt1, prompt2)
        self.assertNotEqual(prompt1, prompt3)
        self.assertNotEqual(prompt1, prompt4)


class TestCommands(TestCase):
    infrastructure_class = SQLAlchemyApplication

    def test_command_aggregate(self):
        # Create a command.
        cmd = Command.__create__()
        assert isinstance(cmd, Command)

        # Check is_done gets changed by done()
        self.assertFalse(cmd.is_done)
        cmd.done()
        self.assertTrue(cmd.is_done)

        # Check the resulting domain events.
        pending_events = cmd.__batch_pending_events__()
        self.assertIsInstance(pending_events[0], Command.Created)
        self.assertIsInstance(pending_events[1], Command.Done)

    def test_command_process(self):
        commands = CommandProcess.mixin(self.infrastructure_class)(setup_table=True)
        core = ProcessApplication.mixin(self.infrastructure_class)(
            "core", policy=example_policy, session=commands.session
        )

        self.assertFalse(list(commands.event_store.all_events()))

        example_id = uuid4()
        cmd = CreateExample.__create__(example_id=example_id)
        # cmd = Command.__create__(cmd_method='create_example', cmd_args={})
        cmd.__save__()

        domain_events = list(commands.event_store.all_events())
        # self.assertTrue(domain_events)
        self.assertEqual(len(domain_events), 1)

        self.assertFalse(list(core.event_store.all_events()))

        core.follow("commands", commands.notification_log)
        core.run()

        self.assertTrue(list(core.event_store.all_events()))
        self.assertIn(example_id, core.repository)

        # Example shouldn't be "moved on" because core isn't following itself,
        # or applying its policy to generated events.
        self.assertFalse(core.repository[example_id].is_moved_on)

        commands.close()
        core.close()

    def test_apply_policy_to_generated_domain_events(self):
        commands = CommandProcess.mixin(self.infrastructure_class)(setup_table=True)
        core = ProcessApplication.mixin(self.infrastructure_class)(
            name="core",
            policy=example_policy,
            session=commands.session,
            apply_policy_to_generated_events=True,
        )

        with core, commands:
            self.assertFalse(list(commands.event_store.all_events()))

            example_id = uuid4()
            cmd = CreateExample.__create__(example_id=example_id)
            cmd.__save__()

            domain_events = list(commands.event_store.all_events())
            self.assertEqual(len(domain_events), 1)

            self.assertFalse(list(core.event_store.all_events()))

            core.follow("commands", commands.notification_log)
            core.run()

            self.assertTrue(list(core.event_store.all_events()))
            self.assertIn(example_id, core.repository)

            # Example should be "moved on" because core is
            # applying its policy to generated events.
            example = core.repository[example_id]
            self.assertTrue(example.is_moved_on)

            # Check the "second" aggregate exists.
            self.assertIn(example.second_id, core.repository)

    def tearDown(self):
        assert_event_handlers_empty()


# Example aggregate (used in the test).


class ExampleAggregate(BaseAggregateRoot):
    def __init__(self, **kwargs):
        super(ExampleAggregate, self).__init__(**kwargs)
        self.is_moved_on = False
        self.second_id = None

    class Event(BaseAggregateRoot.Event[TEntity]):
        pass

    class Created(Event[TEntity], BaseAggregateRoot.Created[TEntity]):
        pass

    def move_on(self, second_id=None):
        self.__trigger_event__(ExampleAggregate.MovedOn, second_id=second_id)

    class MovedOn(Event[TEntity]):
        @property
        def second_id(self):
            return self.__dict__["second_id"]

        def mutate(self, aggregate):
            assert isinstance(aggregate, ExampleAggregate)
            aggregate.is_moved_on = True
            aggregate.second_id = self.second_id


def example_policy(repository, event):
    if isinstance(event, ExampleAggregate.Created):
        # Create a second aggregate, allowing test to check that
        # events from more than one entity are stored.
        second_id = uuid4()
        second = AggregateRoot.__create__(originator_id=second_id)

        # Get first aggregate and move it on, allowing test to
        # check that the first aggregate was mutated by the policy.
        first = repository[event.originator_id]
        assert isinstance(first, ExampleAggregate)
        first.move_on(second_id=second_id)

        return second

    elif isinstance(event, Command.Created):
        command_class = resolve_topic(event.originator_topic)
        if command_class is CreateExample:
            return ExampleAggregate.__create__(event.example_id)


class LogMessage(BaseAggregateRoot):
    def __init__(self, message="", **kwargs):
        super(LogMessage, self).__init__(**kwargs)
        self.message = message

    class Event(BaseAggregateRoot.Event[TEntity]):
        pass

    class Created(Event[TEntity], BaseAggregateRoot.Created[TEntity]):
        pass


def event_logging_policy(_, event):
    return LogMessage.__create__(uuid4(), message=str(event))


class CreateExample(Command):
    def __init__(self, example_id=None, **kwargs):
        super().__init__(**kwargs)
        self.example_id = example_id


class SaveOrmObject(Command):
    pass


class DeleteOrmObject(Command):
    pass



from eventsourcing.tests.datastore_tests.test_cassandra import (
    CassandraDatastoreTestCase,
)
from eventsourcing.tests.sequenced_item_tests import base


class TestCassandraRecordManagerWithIntegerSequences(
    CassandraDatastoreTestCase, base.IntegerSequencedRecordTestCase
):
    pass


class TestCassandraRecordManagerWithTimestampSequences(
    CassandraDatastoreTestCase, base.TimestampSequencedItemTestCase
):
    pass


class WithCassandraRecordManagers(CassandraDatastoreTestCase, base.WithRecordManagers):
    pass


class TestSequencedItemIteratorWithCassandra(
    WithCassandraRecordManagers, base.SequencedItemIteratorTestCase
):
    pass


class TestThreadedSequencedItemIteratorWithCassandra(
    WithCassandraRecordManagers, base.ThreadedSequencedItemIteratorTestCase
):
    pass



import os
import platform

from setuptools import find_packages, setup

from eventsourcing import __version__

is_pypy = platform.python_implementation() == "PyPy"

# Read the docs doesn't need to build the Cassandra driver (and can't).
if "READTHEDOCS" in os.environ:
    os.environ["CASS_DRIVER_NO_CYTHON"] = "1"

install_requires = [
    "python-dateutil<=2.8.99999",
    "pycryptodome<=3.9.99999",
    "requests<=2.23.99999",
    "readerwriterlock<=1.0.99999",
]

sqlalchemy_requires = ["sqlalchemy<=1.3.99999,>=0.9", "sqlalchemy-utils<=0.36.99999"]

axonserver_requires = ["axonclient<=0.0.99999"]

ray_requires = ["ray<=0.8.5", "psutil", "setproctitle"]

thespian_requires = ["thespian<=3.10.99999"]

cassandra_requires = ["cassandra-driver<=3.23.99999"]

django_requires = ["django<=3.0.99999"]

testing_requires = (
    cassandra_requires
    + sqlalchemy_requires
    + axonserver_requires
    + thespian_requires
    + ray_requires
    + django_requires
    + [
        "mock<=4.0.99999",
        "flask<=1.1.99999",
        "flask_sqlalchemy<=2.4.99",
        "uwsgi<=2.0.99999",
        # "redis<=3.5.99999",
        "celery<=4.4.99999",
        "pymysql<=0.9.99999",
        "mysql-connector-python-rf<=2.2.99999",
        "python-coveralls<=2.9.99999",
        "coverage<5.0.0",  # v5 is incompatible ATM.
        "cryptography",
        # Tests use Django with PostgreSQL.
        "psycopg2cffi<=2.8.99999" if is_pypy else "psycopg2-binary<=2.8.99999",
    ]
)

docs_requires = testing_requires + [
    "Sphinx==1.8.5",
    "python_docs_theme",
    "sphinx_py3doc_enhanced_theme",
    "sphinx_rtd_theme==0.4.3",
    "Alabaster",
    "sphinx-autobuild",
]

dev_requires = docs_requires + ["black", "mypy", "flake8", "flake8-bugbear", "isort"]

long_description = """
A library for event sourcing in Python.

`Package documentation is now available <http://eventsourcing.readthedocs.io/>`_.

`Please raise issues on GitHub <https://github.com/johnbywater/eventsourcing/issues>`_.
"""

packages = find_packages(
    exclude=[
        "docs",
        # "eventsourcing.contrib*",
        # "eventsourcing.tests*"
    ]
)

setup(
    name="eventsourcing",
    version=__version__,
    description="Event sourcing in Python",
    author="John Bywater",
    author_email="john.bywater@appropriatesoftware.net",
    url="https://github.com/johnbywater/eventsourcing",
    license="BSD-3-Clause",
    packages=packages,
    package_data={"eventsourcing": ["py.typed"]},
    install_requires=install_requires,
    extras_require={
        "cassandra": cassandra_requires,
        "sqlalchemy": sqlalchemy_requires,
        "axonserver": axonserver_requires,
        "axon": axonserver_requires,
        "ray": ray_requires,
        "django": django_requires,
        "test": testing_requires,
        "tests": testing_requires,
        "testing": testing_requires,
        "dev": dev_requires,
        "docs": docs_requires,
    },
    zip_safe=False,
    long_description=long_description,
    keywords=[
        "event sourcing",
        "event store",
        "domain driven design",
        "ddd",
        "cqrs",
        "cqs",
    ],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        # 'Programming Language :: Python :: 3.5',   # we use f-strings
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)



from json import JSONDecodeError, JSONDecoder
from typing import Optional, Type

import requests

from eventsourcing.application.notificationlog import (
    AbstractNotificationLog,
    LocalNotificationLog,
    Section,
)
from eventsourcing.utils.transcoding import ObjectJSONDecoder, ObjectJSONEncoder


class RemoteNotificationLog(AbstractNotificationLog):
    """
    Presents notification log sections retrieved an HTTP API
    that presents notification log sections in JSON format,
    for example by using a NotificationLogView.
    """

    def __init__(
        self, base_url: str, json_decoder_class: Optional[Type[JSONDecoder]] = None
    ):
        """
        Initialises remote notification log object.

        :param str base_url: A URL for the HTTP API.
        :param JSONDecoder json_decoder_class: used to deserialize remote sections.
        """
        self.base_url = base_url
        json_decoder_class_ = json_decoder_class or ObjectJSONDecoder
        self.json_decoder = json_decoder_class_()
        self._section_size = -1

    def json_loads(self, value: str) -> object:
        try:
            return self.json_decoder.decode(value)
        except JSONDecodeError:
            raise ValueError("Couldn't load JSON string: {}".format(value))

    @property
    def section_size(self) -> int:
        """
        Size of section of notification log.
        """
        if self._section_size == -1:
            resource = self.get_resource(self.make_notification_log_url("section_size"))
            self._section_size = self.deserialize_section_size(resource)
        return self._section_size

    def deserialize_section_size(self, resource: str) -> int:
        try:
            section_size = self.json_loads(resource)
        except:
            raise ValueError(
                "Failed to decode JSON 'section_size' resource: {}".format(resource)
            )
        else:
            if not isinstance(section_size, int):
                raise TypeError(
                    "Section size is not an int: {}".format(type(section_size))
                )
            return section_size

    def __getitem__(self, section_id: str) -> Section:
        """
        Returns a section of notification log.

        :param str section_id: ID of the section of the notification log.
        :return: Identified section of notification log.
        :rtype: Section
        """
        section_json = self.get_json(section_id)
        return self.deserialize_section(section_json)

    def deserialize_section(self, section_json: str) -> Section:
        try:
            obj = self.json_loads(section_json)
            assert isinstance(obj, dict)
            section = Section(**obj)
        except ValueError as e:
            raise ValueError(
                "Couldn't deserialize notification log section: "
                "{}: {}".format(e, section_json)
            )
        return section

    def get_json(self, section_id: str) -> str:
        notification_log_url = self.make_notification_log_url(section_id)
        return self.get_resource(notification_log_url)

    def make_notification_log_url(self, section_id: str) -> str:
        return "{}/{}/".format(self.base_url.strip("/"), section_id)

    def get_resource(self, url: str) -> str:
        return requests.get(url).content.decode("utf8")


class NotificationLogView(object):
    """
    Presents sections of a notification log in JSON format.

    Can be used to make an HTTP API that can be used
    remotely, for example by a RemoteNotificationLog.
    """

    def __init__(
        self, notification_log: LocalNotificationLog, json_encoder: ObjectJSONEncoder
    ):
        """
        Initialises notification log view object.

        :param notification_log: A notification log object
        :param json_encoder_class: JSON encoder class
        """
        assert isinstance(notification_log, LocalNotificationLog), type(
            notification_log
        )
        self.notification_log = notification_log
        self.json_encoder = json_encoder

    def present_resource(self, name: str) -> bytes:
        """
        Returns a resource of the notification log in JSON format.

        Supports returning section of the notification log.

        Also supports returning section_size of notification
        log, if section_id has special value 'section_size'.

        :param name: Name of the resource, e.g. a section ID.
        :return: Identified resource of notification log view in JSON format.
        """
        if name == "section_size":
            # Present the notification log's configured section size.
            section_size = self.notification_log.section_size
            resource = self.json_encoder.encode(section_size)
        else:
            # Default to assuming the resource is a section.
            section = self.notification_log[name]
            resource = self.json_encoder.encode(section.__dict__)

        return resource



from unittest.case import TestCase
from uuid import uuid4

from mock import mock

from eventsourcing.application.decorators import applicationpolicy, applicationpolicy2
from eventsourcing.domain.model.decorators import mutator, retry, subscribe_to
from eventsourcing.domain.model.events import (
    EventHandlersNotEmptyError,
    assert_event_handlers_empty,
    clear_event_handlers,
    publish,
)
from eventsourcing.example.domainmodel import Example
from eventsourcing.utils.topic import get_topic


class TestDecorators(TestCase):
    def tearDown(self):
        clear_event_handlers()

    def test_retry_without_arg(self):
        # Check docstrings of decorated functions.
        def func(*args):
            """func docstring"""
            return 1

        def func_raises(exc=Exception):
            """func_raises docstring"""
            raise exc

        self.assertEqual(retry(func).__doc__, func.__doc__)
        self.assertEqual(retry()(func).__doc__, func.__doc__)
        self.assertEqual(retry(func_raises).__doc__, func_raises.__doc__)
        self.assertEqual(retry()(func_raises).__doc__, func_raises.__doc__)

        # Check func returns correctly after it is decorated.
        self.assertEqual(retry(func)(), 1)
        self.assertEqual(retry()(func)(), 1)

        # Check func raises correctly after it is decorated.
        with self.assertRaises(Exception):
            retry(func_raises, wait=0)()
        with self.assertRaises(Exception):
            retry(wait=0)(func_raises)()

        with self.assertRaises(NotImplementedError):
            retry(func_raises, wait=0)(NotImplementedError)
        with self.assertRaises(NotImplementedError):
            retry(wait=0)(func_raises)(NotImplementedError)

        with self.assertRaises(NotImplementedError):
            retry(func_raises, wait=0, max_attempts=0)(NotImplementedError)
        with self.assertRaises(NotImplementedError):
            retry(wait=0, max_attempts=0)(func_raises)(NotImplementedError)

        # Check TypeError is raised if args have incorrect values.
        with self.assertRaises(TypeError):
            retry(max_attempts="")  # needs to be an int

        with self.assertRaises(TypeError):
            retry(exc=TestCase)  # need to be subclass of Exception (single value)

        with self.assertRaises(TypeError):
            retry(exc=(TestCase,))  # need to be subclass of Exception (tuple)

        with self.assertRaises(TypeError):
            retry(wait="")  # needs to be a float

    def test_mutate_without_arg(self):
        # Define an entity class, and event class, and a mutator function.

        class Entity(object):
            pass

        class Event(object):
            pass

        @mutator
        def mutate_entity(initial, event):
            """Docstring for function mutate()."""
            raise NotImplementedError()

        # Check the function looks like the original.
        self.assertEqual(mutate_entity.__doc__, "Docstring for function mutate().")

        # Check it is capable of registering event handlers.
        @mutate_entity.register(Event)
        def mutate_with_event(initial, event):
            return Entity()

        # Check it dispatches on type of last arg.
        self.assertIsInstance(mutate_entity(None, Event()), Entity)

        # Check it handles unregistered types.
        with self.assertRaises(NotImplementedError):
            mutate_entity(None, None)

    def test_mutate_with_arg(self):
        # Define an entity class, and event class, and a mutator function.

        class Entity(object):
            pass

        class Event(object):
            pass

        @mutator(Entity)
        def mutate_entity(initial, event):
            """Docstring for function mutate()."""
            raise NotImplementedError()

        # Check the function looks like the original.
        self.assertEqual(mutate_entity.__doc__, "Docstring for function mutate().")

        # Check it is capable of registering event handlers.
        @mutate_entity.register(Event)
        def mutate_with_event(initial, event):
            return initial()

        # Check it dispatches on type of last arg.
        self.assertIsInstance(mutate_entity(None, Event()), Entity)

        # Check it handles unregistered types.
        with self.assertRaises(NotImplementedError):
            mutate_entity(None, None)

    def test_subscribe_to_decorator(self):
        entity_id1 = uuid4()
        event1 = Example.Created(
            originator_id=entity_id1, originator_topic=get_topic(Example), a=1, b=2
        )
        event2 = Example.Discarded(originator_id=entity_id1, originator_version=1)
        handler1 = mock.Mock()
        handler2 = mock.Mock()
        handler3 = mock.Mock()

        # Check we can assert there are no event handlers subscribed.
        assert_event_handlers_empty()

        # Original style (one event class arg).
        @subscribe_to(Example.Created)
        def test_handler1(e):
            """Doc string"""
            handler1(e)

        # Naked style (not called).
        @subscribe_to
        def test_handler2(e):
            """Doc string"""
            handler2(e)

        # Multi-event style (many event class args).
        @subscribe_to((Example.Created, Example.Discarded))
        def test_handler3(e):
            """Doc string"""
            handler3(e)

        # Check the decorator doesn't mess with the function doc string.
        self.assertEqual("Doc string", test_handler1.__doc__)
        self.assertEqual("Doc string", test_handler2.__doc__)
        self.assertEqual("Doc string", test_handler3.__doc__)

        # Check can fail to assert event handlers empty.
        self.assertRaises(EventHandlersNotEmptyError, assert_event_handlers_empty)

        # Check event is received when published individually.
        publish([event1])
        handler1.assert_called_once_with(event1)
        handler2.assert_called_once_with(event1)
        handler3.assert_called_once_with(event1)

        # Check event of wrong type is not received.
        handler1.reset_mock()
        handler2.reset_mock()
        handler3.reset_mock()
        publish([event2])
        self.assertFalse(handler1.call_count)
        handler2.assert_called_once_with(event2)
        handler3.assert_called_once_with(event2)

        # Check a list of events can be filtered.
        handler1.reset_mock()
        handler2.reset_mock()
        handler3.reset_mock()
        publish([event1, event2])
        handler1.assert_called_once_with(event1)
        self.assertEqual(handler2.call_count, 2)
        self.assertEqual(handler3.call_count, 2)

        handler1.reset_mock()
        handler2.reset_mock()
        handler3.reset_mock()
        publish([event1, event1])
        self.assertEqual(2, handler1.call_count)
        self.assertEqual(2, handler2.call_count)
        self.assertEqual(2, handler3.call_count)

        handler1.reset_mock()
        handler2.reset_mock()
        handler3.reset_mock()
        publish([event2, event2])
        self.assertEqual(0, handler1.call_count)
        self.assertEqual(2, handler2.call_count)
        self.assertEqual(2, handler3.call_count)

        clear_event_handlers()

    def test_applicationpolicy_decorator(self):

        self.seen_default = False
        self.seen_int = False

        class Application(object):
            def __init__(self, test_case):
                self.test_case = test_case

            @applicationpolicy
            def policy(self, repository, event):
                self.test_case.seen_default = True

            @policy.register(int)
            def _(self, repository, event):
                self.test_case.seen_int = True

        app = Application(self)
        self.assertFalse(self.seen_default)
        self.assertFalse(self.seen_int)
        app.policy(None, "")
        self.assertTrue(self.seen_default)
        self.assertFalse(self.seen_int)
        app.policy(None, 1)
        self.assertTrue(self.seen_default)
        self.assertTrue(self.seen_int)

    def test_applicationpolicy2_decorator(self):

        self.seen_default = False
        self.seen_int = False
        self.seen_a = False
        self.seen_b = False

        class A(object):
            pass

        class B(A):
            pass

        class C(B):
            pass

        class Application(object):
            def __init__(self, test_case):
                self.test_case = test_case

            @applicationpolicy2
            def policy(self, repository, event):
                self.test_case.seen_default = True

            @policy.register(int)
            def _(self, repository, event):
                self.test_case.seen_int = True

            @policy.register(A)
            def _(self, repository, event):
                self.test_case.seen_a = True

            @policy.register(B)
            def _(self, repository, event):
                self.test_case.seen_b = True

        app = Application(self)
        self.assertFalse(self.seen_default)
        self.assertFalse(self.seen_int)
        app.policy(None, "")
        self.assertTrue(self.seen_default)
        self.assertFalse(self.seen_int)
        app.policy(None, 1)
        self.assertTrue(self.seen_default)
        self.assertTrue(self.seen_int)
        app.policy(None, C())
        self.assertTrue(self.seen_b)



from eventsourcing.application.simple import ApplicationWithConcreteInfrastructure
from eventsourcing.infrastructure.popo.factory import PopoInfrastructureFactory
from eventsourcing.infrastructure.popo.mapper import SequencedItemMapperForPopo
from eventsourcing.infrastructure.popo.records import StoredEventRecord


class PopoApplication(ApplicationWithConcreteInfrastructure):
    infrastructure_factory_class = PopoInfrastructureFactory
    sequenced_item_mapper_class = SequencedItemMapperForPopo
    stored_event_record_class = StoredEventRecord



from abc import abstractmethod
from threading import Lock


class AbstractIntegerSequenceGenerator(object):
    """
    Abstract base class for generating a sequence of integers.
    """

    def __iter__(self) -> "AbstractIntegerSequenceGenerator":
        return self

    @abstractmethod
    def __next__(self) -> int:
        """
        Returns the next item in the container.
        """


class SimpleIntegerSequenceGenerator(AbstractIntegerSequenceGenerator):
    """
    Generates a sequence of integers, by simply incrementing a Python int.
    """

    def __init__(self, i: int = 0):
        self.i = i
        self.lock = Lock()

    def __next__(self) -> int:
        """
        Returns the next item in the container.
        """
        self.lock.acquire()
        i = self.i
        self.i += 1
        self.lock.release()
        return i



from unittest.case import TestCase
from uuid import uuid4

from cassandra import InvalidRequest
from cassandra.cluster import NoHostAvailable
from cassandra.cqlengine import CQLEngineException

from eventsourcing.exceptions import DatasourceSettingsError
from eventsourcing.infrastructure.cassandra.datastore import (
    CassandraDatastore,
    CassandraSettings,
)
from eventsourcing.infrastructure.cassandra.factory import (
    CassandraInfrastructureFactory,
)
from eventsourcing.infrastructure.cassandra.records import (
    IntegerSequencedRecord,
    SnapshotRecord,
    TimestampSequencedRecord,
)
from eventsourcing.infrastructure.datastore import (
    DatastoreConnectionError,
    DatastoreTableError,
)
from eventsourcing.tests.datastore_tests import base

DEFAULT_KEYSPACE_FOR_TESTING = "eventsourcing_tests"


class CassandraDatastoreTestCase(base.AbstractDatastoreTestCase):
    """
    Uses the datastore object to set up connection to and tables in Cassandra.
    """

    infrastructure_factory_class = CassandraInfrastructureFactory

    def construct_datastore(self):
        return CassandraDatastore(
            settings=CassandraSettings(default_keyspace=DEFAULT_KEYSPACE_FOR_TESTING),
            tables=(IntegerSequencedRecord, TimestampSequencedRecord, SnapshotRecord),
        )


class TestCassandraDatastore(CassandraDatastoreTestCase, base.DatastoreTestCase):
    def list_records(self):
        try:
            return list(IntegerSequencedRecord.objects.all())
        except (CQLEngineException, NoHostAvailable) as e:
            raise DatastoreConnectionError(e)
        except InvalidRequest as e:
            raise DatastoreTableError(e)

    def create_record(self):
        record = IntegerSequencedRecord(
            sequence_id=uuid4(), position=0, topic="topic", state=b"{}"
        )
        try:
            record.save()
        except (CQLEngineException, NoHostAvailable) as e:
            raise DatastoreConnectionError(e)
        except InvalidRequest as e:
            raise DatastoreTableError(e)
        return record


class TestPlainTextAuthProvider(TestCase):
    def test_plain_text_auth_provider(self):
        datastore = CassandraDatastore(
            settings=CassandraSettings(
                default_keyspace=DEFAULT_KEYSPACE_FOR_TESTING,
                username="username",
                password="password",
            ),
            tables=(IntegerSequencedRecord,),
        )
        datastore.setup_connection()


class TestDatabaseSettingsError(TestCase):
    def test_consistency_level_error(self):
        datastore = CassandraDatastore(
            settings=CassandraSettings(
                default_keyspace=DEFAULT_KEYSPACE_FOR_TESTING, consistency="invalid"
            ),
            tables=(IntegerSequencedRecord,),
        )

        with self.assertRaises(DatasourceSettingsError):
            datastore.setup_connection()



import datetime
import time
from uuid import uuid4

from eventsourcing.domain.model.decorators import attribute
from eventsourcing.domain.model.entity import (
    TimestampedVersionedEntity,
    VersionedEntity,
)
from eventsourcing.domain.model.events import publish, subscribe, unsubscribe
from eventsourcing.example.domainmodel import Example, create_new_example
from eventsourcing.example.infrastructure import ExampleRepository
from eventsourcing.exceptions import (
    ConcurrencyError,
    OriginatorIDError,
    OriginatorVersionError,
    ProgrammingError,
    RepositoryKeyError,
)
from eventsourcing.tests.sequenced_item_tests.base import WithEventPersistence
from eventsourcing.tests.sequenced_item_tests.test_cassandra_record_manager import (
    WithCassandraRecordManagers,
)
from eventsourcing.tests.sequenced_item_tests.test_sqlalchemy_record_manager import (
    SQLAlchemyRecordManagerTestCase,
)
from eventsourcing.utils.times import datetime_from_timestamp
from eventsourcing.utils.topic import get_topic


class TestExampleEntity(SQLAlchemyRecordManagerTestCase, WithEventPersistence):
    def test_entity_lifecycle(self):
        # Check the factory creates an instance.
        time_before = time.time()
        example1 = Example.__create__(a=1, b=2)

        self.assertIsInstance(example1, Example)

        # Check the instance is equal to itself.
        self.assertEqual(example1, example1)

        # Check the instance is equal to a clone of itself.
        clone = object.__new__(type(example1))
        clone.__dict__.update(example1.__dict__)
        self.assertEqual(example1, clone)

        # Check the properties of the Example class.
        self.assertEqual(1, example1.a)
        self.assertEqual(2, example1.b)

        # Check the properties of the TimestampedVersionedEntity class.
        self.assertTrue(example1.id)
        self.assertEqual(example1.__version__, 0)
        self.assertTrue(example1.__created_on__)
        self.assertLess(example1.__created_on__, time.time())
        self.assertGreater(example1.__created_on__, time_before)
        self.assertTrue(example1.__last_modified__)
        self.assertEqual(example1.__created_on__, example1.__last_modified__)

        # Test the datetime_from_timestamp() converts ok.
        tt = time.time()
        dt = datetime.datetime.utcnow()
        self.assertEqual(datetime_from_timestamp(tt).hour, dt.hour)

        # Check can get datetime from timestamps, and it corresponds to UTC.
        dt = datetime_from_timestamp(example1.__created_on__)
        self.assertLess(dt, datetime.datetime.utcnow())
        self.assertGreater(dt, datetime.datetime.utcnow() - datetime.timedelta(1))

        # Check a different type with the same values is not "equal" to the first.
        class Subclass(Example):
            pass

        other = object.__new__(Subclass)
        other.__dict__.update(example1.__dict__)
        self.assertEqual(example1.__dict__, other.__dict__)
        self.assertNotEqual(example1, other)

        # Check a second instance with the same values is not "equal" to the first.
        example2 = create_new_example(a=1, b=2)
        self.assertEqual(type(example1), type(example2))
        self.assertNotEqual(example1, example2)

        # Check entity not hashable.
        with self.assertRaises(TypeError):
            hash(example1)

        # Setup the repo.
        repo = ExampleRepository(self.entity_event_store)

        # Check the example entities can be retrieved from the example repository.
        entity1 = repo[example1.id]
        self.assertIsInstance(entity1, Example)
        self.assertEqual(1, entity1.a)
        self.assertEqual(2, entity1.b)

        entity2 = repo[example2.id]
        self.assertIsInstance(entity2, Example)
        self.assertEqual(1, entity2.a)
        self.assertEqual(2, entity2.b)

        # Check the entity can be updated.
        entity1.a = 100
        self.assertEqual(100, repo[entity1.id].a)
        entity1.b = -200
        self.assertEqual(-200, repo[entity1.id].b)

        self.assertEqual(repo[entity1.id].__created_on__, entity1.__created_on__)
        self.assertEqual(repo[entity1.id].__last_modified__, entity1.__last_modified__)
        self.assertNotEqual(entity1.__last_modified__, entity1.__created_on__)

        self.assertEqual(0, entity1.count_heartbeats())
        entity1.beat_heart()
        entity1.beat_heart()
        entity1.beat_heart()
        self.assertEqual(3, entity1.count_heartbeats())
        self.assertEqual(3, repo[entity1.id].count_heartbeats())

        # Check the entity can publish multiple events simultaneously.
        entity1.beat_heart(number_of_beats=3)
        self.assertEqual(6, repo[entity1.id].count_heartbeats())

        # Check the entity can be discarded.
        entity1.__discard__()

        # Check the repo now raises a KeyError.
        self.assertRaises(RepositoryKeyError, repo.__getitem__, entity1.id)

        # Check the entity can't be discarded twice.
        self.assertRaises(AssertionError, entity1.__discard__)

        # Should fail to validate event with wrong entity ID.
        with self.assertRaises(OriginatorIDError):
            VersionedEntity.Event(
                originator_id=uuid4(), originator_version=0
            ).__check_obj__(entity2)
        # Should fail to validate event with wrong entity version.
        with self.assertRaises(OriginatorVersionError):
            VersionedEntity.Event(
                originator_id=entity2.id,
                originator_version=0,
                __previous_hash__=entity2.__head__,
            ).__check_obj__(entity2)

        # Should validate event with correct entity ID and version.
        VersionedEntity.Event(
            originator_id=entity2.id,
            originator_version=entity2.__version__ + 1,
            __previous_hash__=entity2.__head__,
        ).__check_obj__(entity2)

        # Check an entity cannot be reregistered with the ID of a discarded entity.
        replacement_event = Example.Created(
            originator_id=entity1.id, a=11, b=12, originator_topic=get_topic(Example)
        )
        with self.assertRaises(ConcurrencyError):
            publish(events=[replacement_event])

    def test_attribute(self):
        # Check we get an error when called with something other than a function.
        self.assertRaises(ProgrammingError, attribute, "not a getter")
        self.assertRaises(ProgrammingError, attribute, 123)
        self.assertRaises(ProgrammingError, attribute, None)

        # Call the decorator with a function.
        getter = lambda: None
        p = attribute(getter)

        # Check we got a property object.
        self.assertIsInstance(p, property)

        # Check the property object has both setter and getter functions.
        self.assertTrue(p.fset)
        self.assertTrue(p.fget)

        # Pretend we decorated an object.
        entity_id = uuid4()
        o = VersionedEntity(id=entity_id, __version__=0)
        o.__dict__["_<lambda>"] = "value1"

        # Call the property's getter function.
        value = p.fget(o)
        self.assertEqual(value, "value1")

        # Call the property's setter function.
        p.fset(o, "value2")

        # Check the attribute has changed.
        value = p.fget(o)
        self.assertEqual(value, "value2")

        # Check the property's getter function isn't the getter function we passed in.
        self.assertNotEqual(p.fget, getter)

        # Define a class that uses the decorator.
        class Aaa(VersionedEntity):
            "An event sourced entity."

            def __init__(self, a, *args, **kwargs):
                super(Aaa, self).__init__(*args, **kwargs)
                self._a = a

            @attribute
            def a(self):
                "A mutable event sourced property."

        # Instantiate the class and check assigning to the property publishes an
        # event and updates the object state.
        published_events = []

        def receive(x):
            published_events.extend(x)

        subscription = (lambda x: True, receive)
        subscribe(*subscription)
        entity_id = uuid4()
        try:
            aaa = Aaa(id=entity_id, __version__=1, a=1)
            self.assertEqual(aaa.a, 1)
            aaa.a = "value1"
            self.assertEqual(aaa.a, "value1")
        finally:
            unsubscribe(*subscription)

        # Check an event was published.
        self.assertEqual(len(published_events), 1)

        # Check the published event was an AttributeChanged event, with the expected
        # attribute values.
        published_event = published_events[0]
        self.assertIsInstance(published_event, Aaa.AttributeChanged)
        self.assertEqual(published_event.name, "_a")
        self.assertEqual(published_event.value, "value1")
        self.assertTrue(published_event.originator_version, 1)
        self.assertEqual(published_event.originator_id, entity_id)

    def test_event_is_hashable(self):
        event1 = Example.Event(originator_id=1, originator_version=0, timestamp=1)
        event2 = Example.Event(originator_id=1, originator_version=0, timestamp=1)
        event3 = Example.Event(originator_id=1, originator_version=0, timestamp=2)

        # Same type with same values.
        self.assertEqual(event1, event2)
        self.assertEqual(hash(event1), hash(event2))  # Same thing

        # Same type with different values.
        self.assertNotEqual(event1, event3)
        self.assertNotEqual(hash(event1), hash(event3))  # Same thing

        # Different type with same values.
        class Subclass(Example.Event):
            pass

        event4 = Subclass(originator_id=1, originator_version=0, timestamp=1)

        self.assertNotEqual(hash(event1), hash(event4))  # Same thing
        self.assertNotEqual(event1, event4)
        self.assertNotEqual(hash(event1), hash(event4))  # Same thing

    def test_without_dataintegrity(self):

        # Different type with same values.
        class SubclassEvent(TimestampedVersionedEntity.Event):
            __with_data_integrity__ = False

        event = SubclassEvent(originator_id=1, originator_version=0, timestamp=1)
        self.assertFalse(hasattr(event, "__previous_hash__"))
        self.assertFalse(hasattr(event, "__event_hash__"))

        # Check the Python hash still works.
        self.assertIsInstance(hash(event), int)

        class SubclassCreated(TimestampedVersionedEntity.Created):
            __with_data_integrity__ = False

        event = SubclassCreated(originator_id=1, originator_topic="", timestamp=1)
        self.assertFalse(hasattr(event, "__previous_hash__"))

        entity = SubclassEntity.__create__()
        self.assertFalse(hasattr(entity, "__head__"))


class SubclassEntity(TimestampedVersionedEntity):
    class Event(TimestampedVersionedEntity.Event):
        pass

    class Created(Event, TimestampedVersionedEntity.Created):
        pass


class CustomValueObject(object):
    def __init__(self, value):
        self.value = value


class TestExampleEntityWithCassandra(WithCassandraRecordManagers, TestExampleEntity):
    pass



from unittest import skip

from eventsourcing.application.popo import PopoApplication
from eventsourcing.tests.sequenced_item_tests.test_popo_record_manager import (
    PopoTestCase,
)
from eventsourcing.tests.test_process import TestProcessApplication


class TestProcessWithPopos(PopoTestCase, TestProcessApplication):
    infrastructure_class = PopoApplication

    @skip("Popo record manager doesn't support pipelines")
    def test_causal_dependencies(self):
        super(TestProcessWithPopos, self).test_causal_dependencies()

    @skip("Popo record manager doesn't do projections into custom ORM objects")
    def test_projection_into_custom_orm_obj(self):
        super(TestProcessWithPopos, self).test_projection_into_custom_orm_obj()


del TestProcessApplication
